try:
    import streamlit as st
except ImportError:
    print("Please install streamlit - python3 -m pip install streamlit")
    exit(1)

from pathlib import Path
import matplotlib.pyplot as plt


import tempfile
import sys
import mps
import mps_motion


def get_folder_and_files():
    folder = Path(sys.argv[1])
    return folder, [
        f.name
        for f in folder.iterdir()
        if f.suffix in [".nd2", ".tif", ".tiff", ".czi"]
    ]


def video():
    st.title("Video")

    folder, files = get_folder_and_files()
    filename = st.selectbox("Select file", options=files)

    if st.button("Show video"):
        with st.spinner("Loading video"):
            data = mps.MPS(folder / filename)
            with tempfile.NamedTemporaryFile(suffix=".mp4") as f:
                mps.utils.frames2mp4(
                    data.frames.T,
                    path=f.name,
                    framerate=data.framerate,
                )
                st.video(open(f.name, "rb"), format="video/mp4")


def motion_analysis():
    st.title("Motion analysis")

    folder, files = get_folder_and_files()

    filename = st.selectbox("Select file", options=files)

    cols_mech = st.columns(2)
    with cols_mech[0]:
        flow_algorithm = st.selectbox(
            "Optical flow algorithm",
            mps_motion.list_optical_flow_algorithm(),
        )
        reference_frame = st.number_input("Reference frame", 0)
        show_motion_video = st.checkbox("Show video")

        if show_motion_video:
            vector_scale = st.number_input("Vector scale", value=1, min_value=1)
            step = st.number_input("Step", value=16, min_value=1)

    with cols_mech[1]:
        spacing = st.number_input("spacing", value=1, min_value=1)
        scale = st.number_input("Scale", value=0.4, min_value=0.1, max_value=1.0)

    if not st.button("Run motion analysis"):
        return

    print("Run motion analysis")
    with st.spinner("Read data"):
        data = mps.MPS(folder / filename)

    opt_flow = mps_motion.OpticalFlow(
        data,
        flow_algorithm=flow_algorithm,
        reference_frame=reference_frame,
    )

    with st.spinner("Compute displacement"):
        u = opt_flow.get_displacements(scale=scale)

    mech = mps_motion.Mechanics(u=u, t=data.time_stamps)

    with st.spinner("Compute velocity"):
        v = mech.velocity(spacing=spacing)

    with st.spinner("Compute mean norm displacement"):
        u_norm_mean = u.norm().mean().compute()

    with st.spinner("Compute mean norm velocity"):
        v_norm_mean = v.norm().mean()

    fig, ax = plt.subplots(2, 1, sharex=True)
    ax[0].plot(data.time_stamps, u_norm_mean)
    ax[0].set_ylabel("Displacement [\u00B5m]")
    ax[1].plot(data.time_stamps[:-spacing], v_norm_mean)
    ax[1].set_ylabel("Velocity [\u00B5m/s]")
    st.pyplot(fig=fig)

    if show_motion_video:
        with st.spinner("Creating video"):
            scaled_data = data
            if scale < 1.0:
                scaled_data = mps_motion.scaling.resize_data(data, scale=u.scale)

            with tempfile.NamedTemporaryFile(suffix=".mp4") as f:
                mps_motion.visu.quiver_video(
                    scaled_data,
                    u,
                    f.name,
                    step=step,
                    vector_scale=vector_scale,
                )
                st.video(open(f.name, "rb"), format="video/mp4")

    return


if __name__ == "__main__":
    # Page settings
    st.set_page_config(page_title="MPS motion gui")

    # Sidebar settings
    pages = {
        "Motion analysis": motion_analysis,
        "Video": video,
    }

    st.sidebar.title("mps-motion")

    # Radio buttons to select desired option
    page = st.sidebar.radio("Pages", tuple(pages.keys()))

    pages[page]()

    # About
    st.sidebar.markdown(
        """
    - [Source code](https://github.com/ComputationalPhysiology/mps_motion)
    - [Documentation](http://computationalphysiology.github.io/mps_motion)
    """,
    )



import numpy as np
import pytest
from mps_motion import block_matching as bm


@pytest.mark.parametrize(
    "size, block_size, resize, expected_shape, expected_type",
    [
        ((64, 64), 9, False, (7, 7, 2), np.float64),
        ((64, 64), 9, True, (64, 64, 2), np.float64),
    ],
)
def test_flow_shape(size, block_size, resize, expected_shape, expected_type):

    reference_image = 255 * np.random.randint(0, 255, size=size, dtype=np.uint8)
    image = 255 * np.random.randint(0, 255, size=size, dtype=np.uint8)

    flow = bm.flow(
        image=image,
        reference_image=reference_image,
        block_size=block_size,
        resize=resize,
    )
    assert flow.shape == expected_shape
    assert flow.dtype == expected_type


def test_get_displacements():

    size = (64, 64)
    reference_image = 255 * np.random.randint(0, 255, size=size, dtype=np.uint8)
    image = 255 * np.random.randint(0, 255, size=size, dtype=np.uint8)
    frames = [reference_image, image, image]
    u = bm.get_displacements(
        frames=np.array(frames).T,
        reference_image=reference_image,
    )
    assert u.shape == (size[0], size[1], len(frames), 2)



import typing
from pathlib import Path
from unittest import mock

import numpy as np
import pytest
from mps_motion import frame_sequence as fs
from mps_motion import Mechanics
from mps_motion import MPSData
from mps_motion import utils
from scipy.ndimage import geometric_transform

here = Path(__file__).absolute().parent

_TESTFILE_NAME = here.joinpath("../datasets/mps_data.npy").as_posix()


@pytest.fixture
def TEST_FILENAME():
    return _TESTFILE_NAME


def get_func(t, a=0.001, b=0):
    def func(x):
        return (x[0] * (1 - a * np.sin(t)), x[1] * (1 - b * np.sin(t)))

    return func


@pytest.fixture
def test_data():

    N = 10
    a = 0.0
    b = 0.02
    frame = np.load(here.joinpath("../datasets/first_frame.npy"))
    frames = [frame]
    times = np.linspace(0, np.pi, N)
    for t in times[1:]:
        frames.append(geometric_transform(frame, get_func(t, a, b)))

    frames = np.array(frames).T
    info = {
        "num_frames": N,
        "dt": np.diff(times).mean(),
        "time_unit": "ms",
        "um_per_pixel": 3.25,
        "size_x": frames.shape[0],
        "size_y": frames.shape[1],
    }

    return MPSData(frames=frames, time_stamps=times, info=info)


@pytest.fixture
def mech_obj():

    width = 10
    height = 15
    num_time_points = 4

    u = np.zeros((height, width, num_time_points, 2))
    # First time point is zero
    # Second time point has a linar displacement in x
    u[:, :, 1, 0] = np.fromfunction(
        lambda y, x: x / width,
        shape=(height, width),
        dtype=float,
    )
    # Third points have linear displacement in y
    u[:, :, 2, 1] = np.fromfunction(
        lambda y, x: y / height,
        shape=(height, width),
        dtype=float,
    )
    # Forth is linear in both
    u[:, :, 2, 0] = u[:, :, 1, 0]
    u[:, :, 3, 1] = u[:, :, 2, 1]

    return Mechanics(fs.VectorFrameSequence(u), t=1000 * np.arange(4))


@pytest.fixture
def mech_trace_obj():
    traces = np.load(here.joinpath("example_traces.npy"), allow_pickle=True).item()

    mechanics_mock = mock.Mock(spec=Mechanics)
    mechanics_mock.u.norm().mean.return_value = traces["u_norm"]
    mechanics_mock.velocity().norm().mean.return_value = traces["v_norm"]
    mechanics_mock.t = traces["time"]
    return mechanics_mock


class SyntheticTrace(typing.NamedTuple):
    t: np.ndarray
    u: np.ndarray
    v: np.ndarray


@pytest.fixture
def synthetic_trace():
    t = np.arange(0, 1, 0.01)
    u = utils.ca_transient(t)
    v = np.abs(np.diff(u))
    return SyntheticTrace(t=t, u=u, v=v)



import logging
from enum import Enum
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

import ap_features as apf
import dask.array as da
import numpy as np

from . import block_matching
from . import dualtvl1
from . import farneback
from . import frame_sequence as fs
from . import lucas_kanade
from . import scaling
from . import utils


logger = logging.getLogger(__name__)


class FLOW_ALGORITHMS(str, Enum):
    farneback = "farneback"
    dualtvl1 = "dualtvl1"
    lucas_kanade = "lucas_kanade"
    block_matching = "block_matching"


def list_optical_flow_algorithm():
    return FLOW_ALGORITHMS._member_names_


class RefFrames(str, Enum):
    min = "min"
    max = "max"
    median = "median"
    mean = "mean"


class ReferenceFrameError(RuntimeError):
    pass


def _check_algorithm(alg):
    msg = f"Expected flow algorithm to be one of {FLOW_ALGORITHMS._member_names_}, got {alg}"
    if alg not in FLOW_ALGORITHMS._member_names_:
        raise ValueError(msg)


def estimate_referece_image_from_velocity(
    t: np.ndarray,
    v: np.ndarray,
    rel_tol=0.01,
    raise_on_failure: bool = False,
) -> int:

    # Estimate baseline
    background = apf.background.background(x=t, y=v)

    # Find points close to the baseline
    inds = np.isclose(v, background, atol=np.abs(v.max() - v.min()) * rel_tol)

    msg = (
        "Unable to find any values are the baseline. "
        "Please try a smaller tolerance when estimating the reference image"
    )
    if not inds.any():
        # No values are this close to the baseline
        logger.warning(msg)

        if raise_on_failure:
            raise ReferenceFrameError(msg)

        # If not raising, just return the index of the first frame
        return 0

    # Find biggest connected segment on the baseline
    from collections import Counter

    groups = np.diff(inds).cumsum()
    counts = Counter(groups).most_common()
    group_index = 0
    ref_index_at_baseline = False

    while not ref_index_at_baseline:
        try:
            index = counts[group_index][0]
        except IndexError:
            logger.warning(msg)
            if raise_on_failure:
                raise ReferenceFrameError(msg)
            # Just return the index of the first frame
            return 0

        # Choose the value at the center of this segment
        ref_index = int(np.median(np.where(groups == index)[0]))
        ref_index_at_baseline = inds[ref_index]
        group_index += 1

    return ref_index


def get_reference_image(
    reference_frame: Union[float, str, RefFrames],
    frames: np.ndarray,
    time_stamps: Optional[np.ndarray] = None,
    smooth_ref_transition: bool = True,
) -> np.ndarray:
    """Get reference frame

    Parameters
    ----------
    reference_frame : Union[float, str, RefFrames]
        Either a float of string representing the
        time-point of the frame that should be used as reference
    frames : np.ndarray
        The image frames
    time_stamps : Optional[np.ndarray], optional
        The time stamps, by default None
    smooth_ref_transition : bool, optional
        If true, compute the mean frame of the three closest frames, by default True

    Returns
    -------
    np.ndarray
        reference_image

    """
    try:
        reference_time = float(reference_frame)

    except ValueError:
        refs = RefFrames._member_names_
        msg = (
            "Expected reference frame to be an integer or one of "
            f"{refs}, got {reference_frame}"
        )
        if str(reference_frame) not in refs:
            raise ValueError(msg)
        reference_image = getattr(np, str(reference_frame))(frames, axis=2)
    else:
        if time_stamps is None:
            raise ValueError("Please provide time stamps")
        try:
            reference_frame_index = next(
                i for i, t in enumerate(time_stamps) if t >= reference_time
            )
        except StopIteration:
            reference_frame_index = len(time_stamps) - 1

        reference_frame_index = int(min(reference_frame_index, len(time_stamps) - 1))
        # Pick neighbouring index
        if smooth_ref_transition:
            if reference_frame_index == 0:
                reference_image = frames[
                    :,
                    :,
                    reference_frame_index : reference_frame_index + 3,
                ].mean(-1)
            elif reference_frame_index == len(time_stamps) - 1:
                reference_image = frames[
                    :,
                    :,
                    reference_frame_index - 2 : reference_frame_index + 1,
                ].mean(-1)
            else:
                reference_image = frames[
                    :,
                    :,
                    reference_frame_index - 1 : reference_frame_index + 2,
                ].mean(-1)
        else:
            reference_image = frames[:, :, reference_frame_index]

    return reference_image


class OpticalFlow:
    def __init__(
        self,
        data: utils.MPSData,
        flow_algorithm: FLOW_ALGORITHMS = FLOW_ALGORITHMS.farneback,
        filter_options: Optional[Dict[str, Any]] = None,
        data_scale: float = 1.0,
        **options,
    ):
        self.data = data

        self.flow_algorithm = flow_algorithm

        self._handle_algorithm(options)
        options["filter_options"] = filter_options or {}
        self._data_scale = data_scale

    @property
    def data_scale(self) -> float:
        return self._data_scale

    def _handle_algorithm(self, options):
        _check_algorithm(self.flow_algorithm)

        self._get_velocities = None
        if self.flow_algorithm == FLOW_ALGORITHMS.lucas_kanade:
            self._get_displacements = lucas_kanade.get_displacements
            self.options = lucas_kanade.default_options()

        elif self.flow_algorithm == FLOW_ALGORITHMS.block_matching:
            self._get_displacements = block_matching.get_displacements
            self.options = block_matching.default_options()

        elif self.flow_algorithm == FLOW_ALGORITHMS.farneback:
            self._get_displacements = farneback.get_displacements
            self.options = farneback.default_options()
            self._get_velocities = farneback.get_velocities

        elif self.flow_algorithm == FLOW_ALGORITHMS.dualtvl1:
            self._get_displacements = dualtvl1.get_displacements
            self.options = dualtvl1.default_options()

        self.options.update(options)

    def get_displacements(
        self,
        recompute: bool = False,
        unit: str = "um",
        scale: float = 1.0,
        reference_frame: Union[float, str, RefFrames] = 0,
        smooth_ref_transition: bool = True,
        reference_image: Optional[np.ndarray] = None,
    ) -> fs.VectorFrameSequence:
        """Compute motion of all images relative to reference frame

        Parameters
        ----------
        recompute : bool, optional
            If already computed set this to true if you want to
            recomputed, by default False
        unit : str, optional
            Either 'pixels' or 'um', by default "pixels".
            If using 'um' them the MPSData.info has to contain the
            key 'um_per_pixel'.
        scale : float, optional
            If less than 1.0, down-sample images before estimating motion, by default 1.0
        reference_frame: float, str, RefFrames, optional
            A float or string indicating the reference frame to use. If the value
            is a float, it should refer to the time-point of the reference frame to use.
            If you known the exact index to use, then you can use the `reference_image` parameter,
            by default 0
        smooth_ref_transition : bool, optional
            If true, compute the mean frame of the three closest frames, by default True
        reference_image: np.ndarray, optional
            The reference image to use containing the actual pixel values.
            Note that if you provide an argument for this, then the `reference_frame`
            parameter is not used, by default None

        Returns
        -------
        np.ndarray
            The displacements
        """
        assert unit in ["pixels", "um"]

        if scale > 1.0:
            raise ValueError("Cannot have scale larger than 1.0")

        if scale < 1.0:
            data = scaling.resize_data(self.data, scale)
        else:
            data = self.data

        if reference_image is None:
            reference_image = get_reference_image(
                reference_frame,
                data.frames,
                data.time_stamps,
                smooth_ref_transition=smooth_ref_transition,
            )

        if not hasattr(self, "_displacement") or recompute:

            u = self._get_displacements(data.frames, reference_image, **self.options)
            dx = 1

            scale *= self.data_scale

            u /= scale

            if unit == "um":
                u *= data.info.get("um_per_pixel", 1.0)
                dx *= data.info.get("um_per_pixel", 1.0)
            else:
                u /= scale

            if not isinstance(u, da.Array):
                u = da.from_array(u)

            self._displacement = fs.VectorFrameSequence(u, dx=dx, scale=scale)

        return self._displacement

    def get_velocities(
        self,
        unit: str = "um",
        scale: float = 1.0,
        spacing: int = 1,
    ):
        assert unit in ["pixels", "um"]
        data = self.data

        if scale > 1.0:
            raise ValueError("Cannot have scale larger than 1.0")

        scaled_data = data
        if scale < 1.0:
            scaled_data = scaling.resize_data(data, scale)

        v = self._get_velocities(
            scaled_data.frames, scaled_data.time_stamps, spacing=spacing, **self.options
        )
        dx = 1

        scale *= self.data_scale

        v /= scale

        if unit == "um":
            v *= scaled_data.info.get("um_per_pixel", 1.0)
            dx *= scaled_data.info.get("um_per_pixel", 1.0)
        else:
            v /= scale

        if not isinstance(v, da.Array):
            v = da.from_array(v)

        self._velocity = fs.VectorFrameSequence(v, dx=dx, scale=scale)

        return self._velocity

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"data={self.data}, flow_algorithm={self.flow_algorithm})"
        )



import dask.array as da
import numpy as np
import pytest
from mps_motion import filters


@pytest.mark.parametrize(
    "filter_type, size, sigma",
    [
        (filters.Filters.median, 3, None),
        (filters.Filters.gaussian, None, 1),
    ],
)
def test_filter_vectors_numpy(filter_type, size, sigma):

    shape = (10, 9, 2)
    np.random.seed(1)
    vectors = 10 * np.ones(shape) + np.random.random(shape)

    filtered_vectors = filters.filter_vectors(
        vectors,
        filter_type=filter_type,
        size=size,
        sigma=sigma,
    )
    assert filtered_vectors.shape == shape

    assert 0 < np.abs(filtered_vectors - vectors).max() < 1


@pytest.mark.parametrize(
    "filter_type, size, sigma",
    [
        (filters.Filters.median, None, None),
        (filters.Filters.gaussian, None, None),
        (filters.Filters.median, 0, None),
        (filters.Filters.gaussian, None, 0),
        (filters.Filters.median, None, 3),
        (filters.Filters.gaussian, 1, None),
        ("somefilter", None, None),
    ],
)
def test_filter_with_0_do_nothing(filter_type, size, sigma):
    shape = (10, 9, 2)
    np.random.seed(1)
    vectors = 10 * np.ones(shape) + np.random.random(shape)

    filtered_vectors = filters.filter_vectors(
        vectors,
        filter_type=filter_type,
        size=size,
        sigma=sigma,
    )
    assert filtered_vectors.shape == shape

    assert np.isclose(filtered_vectors, vectors).all()


@pytest.mark.parametrize(
    "filter_type, size, sigma",
    [
        (filters.Filters.median, 3, None),
        (filters.Filters.gaussian, None, 1),
    ],
)
def test_filter_vectors_dask(filter_type, size, sigma):

    shape = (10, 9, 2)
    np.random.seed(1)
    vectors = 10 * da.ones(shape) + da.random.random(shape)

    filtered_vectors = filters.filter_vectors(
        vectors,
        filter_type=filter_type,
        size=size,
        sigma=sigma,
    )
    assert filtered_vectors.shape == shape
    assert 0 < da.absolute(filtered_vectors - vectors).max().compute() < 1


@pytest.mark.parametrize(
    "filter_type, size, sigma",
    [
        (filters.Filters.median, 3, None),
        (filters.Filters.gaussian, None, 1),
    ],
)
def test_filter_vectors_par_numpy(filter_type, size, sigma):

    shape = (10, 10, 5, 2)
    np.random.seed(1)
    vectors = 10 * np.ones(shape) + np.random.random(shape)

    filtered_vectors = filters.filter_vectors_par(
        vectors,
        filter_type=filter_type,
        size=size,
        sigma=sigma,
    )
    assert filtered_vectors.shape == shape
    assert 0 < np.abs(filtered_vectors - vectors).max() < 1



"""
Farnebäck, G. (2003, June). Two-frame motion estimation based on polynomial expansion. In Scandinavian conference on Image analysis (pp. 363-370). Springer, Berlin, Heidelberg.
"""
import logging

import cv2
import dask
import dask.array as da
import numpy as np
from dask.diagnostics import ProgressBar

from . import utils

logger = logging.getLogger(__name__)


def default_options():
    return dict(
        pyr_scale=0.5,
        levels=3,
        winsize=15,
        iterations=3,
        poly_n=5,
        poly_sigma=1.2,
        flags=0,
    )


def flow(
    image: np.ndarray,
    reference_image: np.ndarray,
    pyr_scale: float = 0.5,
    levels: int = 3,
    winsize: int = 15,
    iterations: int = 3,
    poly_n: int = 5,
    poly_sigma: float = 1.2,
    flags: int = 0,
    factor: float = 1.0,
):
    """
    Compute the optical flow using the Farneback method from
    the reference frame to another image

    Parameters
    ----------
    image : np.ndarray
        The target image
    reference_image : np.ndarray
        The reference image
    pyr_scale : float, optional
        parameter, specifying the image scale (<1) to build pyramids
        for each image; pyr_scale=0.5 means a classical pyramid,
        where each next layer is twice smaller than the previous
        one, by default 0.5
    levels : int, optional
        number of pyramid layers including the initial image; levels=1
        means that no extra layers are created and only the original
        images are used, by default 3
    winsize : int, optional
        averaging window size; larger values increase the algorithm
        robustness to image noise and give more chances for fast motion
        detection, but yield more blurred motion field, by default 15
    iterations : int, optional
        number of iterations the algorithm does at each pyramid level, by default 3
    poly_n : int, optional
        size of the pixel neighborhood used to find polynomial expansion in each pixel.
        larger values mean that the image will be approximated with smoother surfaces,
        yielding more robust algorithm and more blurred motion field,
        typically poly_n =5 or 7., by default 5
    poly_sigma : float, optional
        standard deviation of the Gaussian that is used to smooth derivatives used as a
        basis for the polynomial expansion; for poly_n=5, you can set poly_sigma=1.1,
        for poly_n=7, a good value would be poly_sigma=1.5, by default 1.2
    flags : int, optional
        By default 0. operation flags that can be a combination of the following:
        - OPTFLOW_USE_INITIAL_FLOW uses the input flow as an initial flow approximation.
        - OPTFLOW_FARNEBACK_GAUSSIAN uses the Gaussian winsize x winsize filter
        instead of a box filter of the same size for optical flow estimation;
        usually, this option gives z more accurate flow than with a box filter,
        at the cost of lower speed; normally, winsize for a Gaussian window should
        be set to a larger value to achieve the same level of robustness.
    factor: float
        Factor to multiply the result

    Returns
    -------
    np.ndarray
        The motion vectors
    """
    if image.dtype != "uint8":
        image = utils.to_uint8(image)
    if reference_image.dtype != "uint8":
        reference_image = utils.to_uint8(reference_image)

    return factor * cv2.calcOpticalFlowFarneback(
        reference_image,
        image,
        None,
        pyr_scale,
        levels,
        winsize,
        iterations,
        poly_n,
        poly_sigma,
        flags,
    )


def get_displacements(
    frames,
    reference_image: np.ndarray,
    pyr_scale: float = 0.5,
    levels: int = 3,
    winsize: int = 15,
    iterations: int = 3,
    poly_n: int = 5,
    poly_sigma: float = 1.2,
    flags: int = 0,
    **kwargs,
) -> utils.Array:
    """Compute the optical flow using the Farneback method from
    the reference frame to all other frames

    Parameters
    ----------
    frames : np.ndarray
        The frames with some moving objects
    reference_image : np.ndarray
        The reference image
    pyr_scale : float, optional
        parameter, specifying the image scale (<1) to build pyramids
        for each image; pyr_scale=0.5 means a classical pyramid,
        where each next layer is twice smaller than the previous
        one, by default 0.5
    levels : int, optional
        number of pyramid layers including the initial image; levels=1
        means that no extra layers are created and only the original
        images are used, by default 3
    winsize : int, optional
        averaging window size; larger values increase the algorithm
        robustness to image noise and give more chances for fast motion
        detection, but yield more blurred motion field, by default 15
    iterations : int, optional
        number of iterations the algorithm does at each pyramid level, by default 3
    poly_n : int, optional
        size of the pixel neighborhood used to find polynomial expansion in each pixel.
        larger values mean that the image will be approximated with smoother surfaces,
        yielding more robust algorithm and more blurred motion field,
        typically poly_n =5 or 7., by default 5
    poly_sigma : float, optional
        standard deviation of the Gaussian that is used to smooth derivatives used as a
        basis for the polynomial expansion; for poly_n=5, you can set poly_sigma=1.1,
        for poly_n=7, a good value would be poly_sigma=1.5, by default 1.2
    flags : int, optional
        By default 0. operation flags that can be a combination of the following:
        - OPTFLOW_USE_INITIAL_FLOW uses the input flow as an initial flow approximation.
        - OPTFLOW_FARNEBACK_GAUSSIAN uses the Gaussian winsize x winsize filter
        instead of a box filter of the same size for optical flow estimation;
        usually, this option gives z more accurate flow than with a box filter,
        at the cost of lower speed; normally, winsize for a Gaussian window should
        be set to a larger value to achieve the same level of robustness.

    Returns
    -------
    Array
        An array of motion vectors relative to the reference image. If shape of
        input frames are (N, M, T) then the shape of the output is (N, M, T, 2).
    """
    if kwargs:
        logger.warning(f"Unknown arguments {kwargs!r} - ignoring")
    logger.info("Get displacements using Farneback's algorithm")

    all_flows = []
    for im in np.rollaxis(frames, 2):
        all_flows.append(
            dask.delayed(flow)(
                im,
                reference_image,
                pyr_scale,
                levels,
                winsize,
                iterations,
                poly_n,
                poly_sigma,
                flags,
            ),
        )

    with ProgressBar():
        arr = da.compute(all_flows)
    flows = np.stack(*arr, axis=2)  # type:ignore

    logger.info("Done running Farneback's algorithm")
    return da.from_array(flows)


def get_velocities(
    frames: np.ndarray,
    time_stamps: np.ndarray,
    spacing: int = 1,
    pyr_scale: float = 0.5,
    levels: int = 3,
    winsize: int = 15,
    iterations: int = 3,
    poly_n: int = 5,
    poly_sigma: float = 1.2,
    flags: int = 0,
) -> utils.Array:
    """Compute the optical flow using the Farneback method from
    the reference frame to all other frames

    Parameters
    ----------
    frames : np.ndarray
        The frames with some moving objects
    time_stamps : np.ndarray
        Time stamps
    spacing : int
        Spacing between frames used to compute velocities
    pyr_scale : float, optional
        parameter, specifying the image scale (<1) to build pyramids
        for each image; pyr_scale=0.5 means a classical pyramid,
        where each next layer is twice smaller than the previous
        one, by default 0.5
    levels : int, optional
        number of pyramid layers including the initial image; levels=1
        means that no extra layers are created and only the original
        images are used, by default 3
    winsize : int, optional
        averaging window size; larger values increase the algorithm
        robustness to image noise and give more chances for fast motion
        detection, but yield more blurred motion field, by default 15
    iterations : int, optional
        number of iterations the algorithm does at each pyramid level, by default 3
    poly_n : int, optional
        size of the pixel neighborhood used to find polynomial expansion in each pixel.
        larger values mean that the image will be approximated with smoother surfaces,
        yielding more robust algorithm and more blurred motion field,
        typically poly_n =5 or 7., by default 5
    poly_sigma : float, optional
        standard deviation of the Gaussian that is used to smooth derivatives used as a
        basis for the polynomial expansion; for poly_n=5, you can set poly_sigma=1.1,
        for poly_n=7, a good value would be poly_sigma=1.5, by default 1.2
    flags : int, optional
        By default 0. operation flags that can be a combination of the following:
        - OPTFLOW_USE_INITIAL_FLOW uses the input flow as an initial flow approximation.
        - OPTFLOW_FARNEBACK_GAUSSIAN uses the Gaussian winsize x winsize filter
        instead of a box filter of the same size for optical flow estimation;
        usually, this option gives z more accurate flow than with a box filter,
        at the cost of lower speed; normally, winsize for a Gaussian window should
        be set to a larger value to achieve the same level of robustness.

    Returns
    -------
    Array
        An array of motion vectors relative to the reference image. If shape of
        input frames are (N, M, T) then the shape of the output is (N, M, T, 2).
    """

    logger.info("Get velocities using Farneback's algorithm")
    dts = np.subtract(time_stamps[spacing:], time_stamps[:-spacing])
    all_flows = []
    for index in range(frames.shape[-1] - spacing):
        all_flows.append(
            dask.delayed(flow)(
                frames[:, :, index + spacing],
                frames[:, :, index],
                pyr_scale,
                levels,
                winsize,
                iterations,
                poly_n,
                poly_sigma,
                flags,
                (1.0 / dts[index]),
            ),
        )

    with ProgressBar():
        arr = da.compute(all_flows)

    flows = np.stack(*arr, axis=2)  # type:ignore

    logger.info("Done running Farneback's algorithm")

    return da.from_array(flows)



#!/usr/bin/env python3
# c) 2001-2021 Simula Research Laboratory ALL RIGHTS RESERVED
#
# END-USER LICENSE AGREEMENT
# PLEASE READ THIS DOCUMENT CAREFULLY. By installing or using this
# software you agree with the terms and conditions of this license
# agreement. If you do not accept the terms of this license agreement
# you may not install or use this software.
# Permission to use, copy, modify and distribute any part of this
# software for non-profit educational and research purposes, without
# fee, and without a written agreement is hereby granted, provided
# that the above copyright notice, and this license agreement in its
# entirety appear in all copies. Those desiring to use this software
# for commercial purposes should contact Simula Research Laboratory AS:
# post@simula.no
#
# IN NO EVENT SHALL SIMULA RESEARCH LABORATORY BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE
# "MPS" EVEN IF SIMULA RESEARCH LABORATORY HAS BEEN ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE. THE SOFTWARE PROVIDED HEREIN IS
# ON AN "AS IS" BASIS, AND SIMULA RESEARCH LABORATORY HAS NO OBLIGATION
# TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
# SIMULA RESEARCH LABORATORY MAKES NO REPRESENTATIONS AND EXTENDS NO
# WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESSED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
import concurrent.futures
import logging
from typing import Tuple
from typing import Union

import numpy as np
import tqdm

from . import scaling
from . import utils


__author__ = "Henrik Finsberg (henriknf@simula.no), 2017--2020"
__maintainer__ = "Henrik Finsberg"
__email__ = "henriknf@simula.no"


logger = logging.getLogger(__name__)


def default_options():
    return dict(block_size="auto", max_block_movement="auto")


def flow(
    image: np.ndarray,
    reference_image: np.ndarray,
    block_size: Union[str, int] = "auto",
    max_block_movement: Union[str, int] = "auto",
    resize: bool = True,
):
    """
    Computes the displacements from `reference_image` to `image`
    using a block matching algorithm. Briefly, we subdivde the images
    into blocks of size `block_size x block_size` and compare the two images
    within a range of +/- max_block_movement for each block.

    Arguments
    ---------
    image : np.ndarray
        The frame that you want to compute displacement for relative to
        the referernce frame
    reference_image : np.ndarray
        The frame used as reference
    block_size : int
        Size of the blocks
    max_block_movement : int
        Maximum allowed movement of blocks when searching for best match.
    resize: bool
        If True, make sure to resize the output images to have the same
        shape as the input, by default True.

    Note
    ----
    Make sure to have max_block_movement big enough. If this is too small
    then the results will be wrong. It is better to choose a too large value
    of this. However, choosing a too large value will mean that you need to
    compare more blocks which will increase the running time.
    """

    block_size = resolve_block_size(block_size, reference_image.shape)
    max_block_movement = resolve_max_block_movement(max_block_movement, block_size)

    vectors = _flow(reference_image, image, block_size, max_block_movement)

    if resize:
        new_shape: Tuple[int, int] = (
            reference_image.shape[0],
            reference_image.shape[1],
        )
        vectors = scaling.resize_vectors(vectors, new_shape)
    return vectors


@utils.jit(nopython=True)
def _flow(
    image: np.ndarray,
    reference_image: np.ndarray,
    block_size: int = 9,
    max_block_movement: int = 18,
):
    """
    Computes the displacements from `reference_image` to `image`
    using a block matching algorithm. Briefly, we subdivde the images
    into blocks of size `block_size x block_size` and compare the two images
    within a range of +/- max_block_movement for each block.

    Arguments
    ---------
    reference_image : np.ndarray
        The frame used as reference
    image : np.ndarray
        The frame that you want to compute displacement for relative to
        the referernce frame
    block_size : int
        Size of the blocks
    max_block_movement : int
        Maximum allowed movement of blocks when searching for best match.

    Note
    ----
    Make sure to have max_block_movement big enough. If this is too small
    then the results will be wrong. It is better to choose a too large value
    of this. However, choosing a too large value will mean that you need to
    compare more blocks which will increase the running time.
    """
    # Shape of the image that is returned
    y_size, x_size = image.shape
    block_size = max(block_size, 1)
    shape = (max(y_size // block_size, 1), max(x_size // block_size, 1))
    vectors = np.zeros((shape[0], shape[1], 2))
    costs = np.ones((2 * max_block_movement + 1, 2 * max_block_movement + 1))

    # Need to copy images to float array
    # otherwise negative values will be converted to large 16-bit integers
    ref_block = np.zeros((block_size, block_size))  # Block for reference image
    block = np.zeros((block_size, block_size))  # Block for image

    # Loop over each block
    for y_block in range(shape[0]):
        for x_block in range(shape[1]):

            # Coordinates in the orignal imagee
            y_image = y_block * block_size
            x_image = x_block * block_size

            block[:] = image[
                y_image : y_image + block_size,
                x_image : x_image + block_size,
            ]

            # Check if box has values
            if np.max(block) > 0:

                # Loop over values around the block within the `max_block_movement` range
                for i, y_block_ref in enumerate(
                    range(-max_block_movement, max_block_movement + 1),
                ):
                    for j, x_block_ref in enumerate(
                        range(
                            -max_block_movement,
                            max_block_movement + 1,
                        ),
                    ):

                        y_image_ref = y_image + y_block_ref
                        x_image_ref = x_image + x_block_ref

                        # Just make sure that we are within the referece image
                        if (
                            y_image_ref < 0
                            or y_image_ref + block_size > y_size
                            or x_image_ref < 0
                            or x_image_ref + block_size > x_size
                        ):
                            costs[i, j] = np.nan
                        else:
                            ref_block[:] = reference_image[
                                y_image_ref : y_image_ref + block_size,
                                x_image_ref : x_image_ref + block_size,
                            ]
                            # Could improve this cost function / template matching
                            costs[i, j] = np.sum(
                                np.abs(np.subtract(block, ref_block)),
                            ) / (block_size**2)

                # Find minimum cost vector and store it
                dy, dx = np.where(costs == np.nanmin(costs))

                # If there are more then one minima then we select none
                if len(dy) > 1 or len(dx) > 1:
                    vectors[y_block, x_block, 0] = 0
                    vectors[y_block, x_block, 1] = 0
                else:
                    vectors[y_block, x_block, 0] = dx[0] - max_block_movement
                    vectors[y_block, x_block, 1] = dy[0] - max_block_movement
            else:
                # If no values in box set to no movement
                vectors[y_block, x_block, :] = 0

    return vectors


def flow_map(args):
    """
    Helper function for running block maching algorithm in paralell
    """
    return _flow(*args)


def resolve_block_size(block_size: Union[str, int], shape: Tuple[int, ...]) -> int:
    if isinstance(block_size, str):  # == "auto":
        block_size = max(int(min(shape) / 128), 2)
    return block_size


def resolve_max_block_movement(
    max_block_movement: Union[str, int],
    block_size: int,
) -> int:
    if isinstance(max_block_movement, str):  # == "auto":
        max_block_movement = 2 * block_size
    return max_block_movement


def get_displacements(
    frames: np.ndarray,
    reference_image: np.ndarray,
    block_size: Union[str, int] = "auto",
    max_block_movement: Union[str, int] = "auto",
    resize=True,
    **kwargs,
) -> utils.Array:
    """Computes the displacements from `reference_image` to all `frames`
    using a block matching algorithm. Briefly, we subdivde the images
    into blocks of size `block_size x block_size` and compare the two images
    within a range of +/- max_block_movement for each block.

    Arguments
    ---------
    frames : np.ndarray
        The frame that you want to compute displacement for relative to
        the referernce frame. Input must be of shape (N, M, T, 2), where
        N is the width, M is the height and  T is the number
        number of frames.
    reference_image : np.ndarray
        The frame used as reference
    block_size : int
        Size of the blocks, by default 9
    max_block_movement : int
        Maximum allowed movement of blocks when searching for best match,
        by default 18.
    resize: bool
        If True, make sure to resize the output images to have the same
        shape as the input, by default True.

    Note
    ----
    Make sure to have max_block_movement big enough. If this is too small
    then the results will be wrong. It is better to choose a too large value
    of this. However, choosing a too large value will mean that you need to
    compare more blocks which will increase the running time.

    Returns
    -------
    utils.Array
        utils.Array with displacement of shape (N', M', T', 2), where
        N' is the width, M' is the height and  T' is the number
        number of frames. Note if `resize=True` then we have
        (N, M, T, 2) = (N', M', T', 2).
    """
    if kwargs:
        logger.warning(f"Unknown arguments {kwargs!r} - ignoring")
    frames = utils.check_frame_dimensions(frames, reference_image)

    block_size = resolve_block_size(block_size, reference_image.shape)
    max_block_movement = resolve_max_block_movement(max_block_movement, block_size)

    logger.info("Get displacements using block mathching")
    args = (
        (im, reference_image, block_size, max_block_movement)
        for im in np.rollaxis(frames, 2)
    )
    num_frames = frames.shape[-1]

    y_size, x_size = reference_image.shape
    block_size = max(block_size, 1)
    shape = (max(y_size // block_size, 1), max(x_size // block_size, 1))
    flows = np.zeros((shape[0], shape[1], num_frames, 2))

    with concurrent.futures.ProcessPoolExecutor() as executor:
        for i, uv in tqdm.tqdm(
            enumerate(executor.map(flow_map, args)),
            desc="Compute displacement",
            total=num_frames,
        ):
            flows[:, :, i, :] = uv

    if resize:

        new_shape: Tuple[int, int] = (
            reference_image.shape[0],
            reference_image.shape[1],
        )
        flows = scaling.resize_vectors(flows, new_shape)

    return flows



from collections import namedtuple
from pathlib import Path

import mps_motion as mt
import numpy as np
import pytest
from scipy.ndimage import geometric_transform

HERE = Path(__file__).parent.absolute()

SyntData = namedtuple("SyntData", ["data", "mech"])


def get_func(a=0.005, b=0.007):
    def func(x):
        return (x[0] * (1 - a), x[1] * (1 - b))

    return func


@pytest.fixture(scope="session", params=[1, 0.325])
def gen_data(request):
    frame1 = np.load(HERE.joinpath("first_frame.npy"))
    func1 = get_func(a=0.01, b=0.02)
    frame2 = geometric_transform(frame1.T, func1).T
    func2 = get_func(a=0.02, b=0.01)
    frame3 = geometric_transform(frame1.T, func2).T
    frames = np.stack([frame1, frame2, frame3], axis=-1)

    time_stamps = np.array([0, 1, 2])

    info = {
        "size_x": frames.shape[0],
        "size_y": frames.shape[1],
        "num_frames": frames.shape[2],
        "um_per_pixel": request.param,
    }

    data = mt.utils.MPSData(
        frames=frames,
        time_stamps=time_stamps,
        info=info,
        metadata={},
    )
    x, y = np.meshgrid(np.arange(frame1.shape[1]), np.arange(frame1.shape[0]))

    u01, v01 = func1([x, y])
    u_exact1 = x - u01
    v_exact1 = y - v01
    u1 = np.stack([u_exact1, v_exact1], axis=-1)

    u02, v02 = func2([x, y])
    u_exact2 = x - u02
    v_exact2 = y - v02
    u2 = np.stack([u_exact2, v_exact2], axis=-1)

    u0 = np.zeros_like(u1)
    u = mt.VectorFrameSequence(np.stack([u0, u1, u2], axis=2))

    mech = mt.Mechanics(u, time_stamps)

    return SyntData(data=data, mech=mech)


def test_synthetic_strain(gen_data):
    E = gen_data.mech.E
    du = gen_data.mech.du
    F = gen_data.mech.F

    assert (E[:, :, 0, :, :] == 0).all()

    a = 0
    b = 0.02
    c = 0.01
    d = 0

    assert (np.abs(du[:, :, 1, 0, 0] - a) < 1e-12).all()
    assert (np.abs(du[:, :, 1, 0, 1] - b) < 1e-12).all()
    assert (np.abs(du[:, :, 1, 1, 0] - c) < 1e-12).all()
    assert (np.abs(du[:, :, 1, 1, 1] - d) < 1e-12).all()

    assert (np.abs(F[:, :, 1, 0, 0] - (1 + a)) < 1e-12).all()
    assert (np.abs(F[:, :, 1, 0, 1] - b) < 1e-12).all()
    assert (np.abs(F[:, :, 1, 1, 0] - c) < 1e-12).all()
    assert (np.abs(F[:, :, 1, 1, 1] - (1 + d)) < 1e-12).all()

    assert (np.abs(E[:, :, 1, 0, 0] - 0.5 * ((1 + a) ** 2 + c**2 - 1)) < 1e-12).all()
    assert (np.abs(E[:, :, 1, 0, 1] - 0.5 * (b * (1 + a) + c * (1 + d))) < 1e-12).all()
    assert (np.abs(E[:, :, 1, 1, 0] - 0.5 * (b * (1 + a) + c * (1 + d))) < 1e-12).all()
    assert (np.abs(E[:, :, 1, 1, 1] - 0.5 * ((1 + d) ** 2 + b**2 - 1)) < 1e-12).all()


@pytest.mark.parametrize(
    "flow, tol",
    [
        (mt.farneback.flow, 0.05),
        (mt.lucas_kanade.flow, 0.21),
        (mt.block_matching.flow, 0.9),
        (mt.dualtvl1.flow, 0.05),
    ],
)
def test_synthetic_flow(flow, tol, gen_data):
    frame1 = gen_data.data.frames[:, :, 0]
    frame2 = gen_data.data.frames[:, :, 1]

    u_exact = gen_data.mech.u[:, :, 1, 0]
    v_exact = gen_data.mech.u[:, :, 1, 1]

    U = flow(frame2, frame1)

    u = U[:, :, 0]
    v = U[:, :, 1]

    assert np.linalg.norm(u - u_exact) / np.linalg.norm(u_exact) < tol
    assert np.linalg.norm(v - v_exact) / np.linalg.norm(v_exact) < tol


@pytest.mark.parametrize("unit", ["um", "pixels"])
def test_synthetic_pixels(unit, gen_data):
    data = gen_data.data
    u_exact = gen_data.mech.u

    opt_flow = mt.OpticalFlow(
        data,
        mt.FLOW_ALGORITHMS.farneback,
    )

    u = opt_flow.get_displacements(
        recompute=True,
        unit=unit,
        reference_frame=0,
        smooth_ref_transition=False,
    )

    u_exact_norm = u_exact.norm()
    u_norm = u.norm()

    # Exact solution is zero so lets do the mean
    assert np.abs((u_exact_norm - u_norm)[:, :, 0]).mean().compute() < 0.001
    fac = 1 if unit == "pixels" else data.info["um_per_pixel"]

    assert (
        np.linalg.norm((u_exact_norm * fac - u_norm)[:, :, 1])
        / np.linalg.norm(u_exact_norm[:, :, 1] * fac)
    ).compute() < 0.1

    assert (
        np.linalg.norm((u_exact_norm * fac - u_norm)[:, :, 2])
        / np.linalg.norm(u_exact_norm[:, :, 2] * fac)
    ).compute() < 0.1

    mech = mt.Mechanics(u)

    dE = (gen_data.mech.E - mech.E).norm()

    assert np.abs(dE[:, :, 0]).mean().compute() < 0.01
    assert np.abs(dE[:, :, 1]).mean().compute() < 0.05
    assert np.abs(dE[:, :, 2]).mean().compute() < 0.05



import numpy as np
import pytest
from mps_motion import lucas_kanade as lk


@pytest.mark.parametrize(
    "size, step, interpolation, expected_shape, expected_type",
    [
        ((64, 64), 16, "none", (16, 2), np.float32),
        ((64, 64), 16, "reshape", (4, 4, 2), np.float64),
        ((64, 64), 16, "nearest", (64, 64, 2), np.float64),
        ((64, 64), 16, "rbf", (64, 64, 2), np.float64),
    ],
)
def test_flow_shape(size, step, interpolation, expected_shape, expected_type):

    reference_image = 255 * np.random.randint(0, 255, size=size, dtype=np.uint8)
    image = 255 * np.random.randint(0, 255, size=size, dtype=np.uint8)

    reference_points = lk.get_uniform_reference_points(reference_image, step)
    flow = lk.flow(
        image,
        reference_image,
        reference_points,
        interpolation=interpolation,
    )

    assert flow.shape == expected_shape
    assert flow.dtype == expected_type


@pytest.mark.parametrize(
    "input_shape, step",
    [
        ((64, 64), 48),
        ((128, 128), 48),
        ((300, 200), 48),
        ((100, 50), 12),
        ((121, 241), 13),
    ],
)
def test_get_uniform_reference_points(input_shape, step):
    ref_points = lk.get_uniform_reference_points(np.zeros(input_shape), step=step)
    assert ref_points.shape == (
        ((input_shape[0] + step // 2) // step) * ((input_shape[1] + step // 2) // step),
        1,
        2,
    )


def test_get_displacements():

    size = (64, 64)
    reference_image = 255 * np.random.randint(0, 255, size=size, dtype=np.uint8)
    image = 255 * np.random.randint(0, 255, size=size, dtype=np.uint8)
    frames = [reference_image, image, image]
    u = lk.get_displacements(
        frames=np.array(frames).T,
        reference_image=reference_image,
    )
    assert u.shape == (size[0], size[1], len(frames), 2)



"""

ZACH, Christopher; POCK, Thomas; BISCHOF, Horst. A duality based approach for realtime tv-l 1 optical flow. In: Joint pattern recognition symposium. Springer, Berlin, Heidelberg, 2007. p. 214-223.

http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.709.4597&rep=rep1&type=pdf

"""
import logging

import cv2
import dask
import dask.array as da
import numpy as np
from dask.diagnostics import ProgressBar

from . import utils


logger = logging.getLogger(__name__)


def default_options():
    return {"tau": 0.25, "lmbda": 0.08, "theta": 0.37, "nscales": 6, "warps": 5}


def flow(
    image: np.ndarray,
    reference_image: np.ndarray,
    tau: float = 0.25,
    lmbda: float = 0.08,
    theta: float = 0.37,
    nscales: int = 6,
    warps: int = 5,
) -> np.ndarray:

    dual_proc = cv2.optflow.DualTVL1OpticalFlow_create(
        tau,
        lmbda,
        theta,
        nscales,
        warps,
    )
    est_flow = np.zeros(
        shape=(reference_image.shape[0], reference_image.shape[1], 2),
        dtype=np.float32,
    )

    if image.dtype != "uint8":
        image = utils.to_uint8(image)
    if reference_image.dtype != "uint8":
        reference_image = utils.to_uint8(reference_image)

    dual_proc.calc(
        reference_image,
        image,
        est_flow,
    )
    return est_flow


def get_displacements(
    frames: np.ndarray,
    reference_image: np.ndarray,
    tau: float = 0.25,
    lmbda: float = 0.08,
    theta: float = 0.37,
    nscales: int = 6,
    warps: int = 5,
    **kwargs,
) -> utils.Array:
    """Compute the optical flow using the Dual TV-L1 method from
    the reference frame to all other frames


    Parameters
    ----------
    frames : np.ndarray
        The frames with some moving objects
    reference_image : np.ndarray
        The reference image
    tau : float, optional
        Time step of the numerical scheme, by default 0.25
    lmbda : float, optional
        Weight parameter for the data term, attachment parameter.
        This is the most relevant parameter, which determines the smoothness
        of the output. The smaller this parameter is, the smoother the solutions
        we obtain. It depends on the range of motions of the images, so its
        value should be adapted to each image sequence, by default 0.08
    theta : float, optional
        parameter used for motion estimation. It adds a variable allowing for
        illumination variations Set this parameter to 1. if you have varying
        illumination. See: Chambolle et al, A First-Order Primal-Dual Algorithm
        for Convex Problems with Applications to Imaging Journal of Mathematical
        imaging and vision, may 2011 Vol 40 issue 1, pp 120-145, by default 0.37
    nscales : int, optional
        Number of scales used to create the pyramid of images, by default 6
    warps : int, optional
        Number of warpings per scale. Represents the number of times that I1(x+u0)
        and grad( I1(x+u0) ) are computed per scale. This is a parameter that assures
        the stability of the method. It also affects the running time, so it is a
        compromise between speed and accuracy., by default 5

    Returns
    -------
    Array
        An array of motion vectors relative to the reference image. If shape of
        input frames are (N, M, T) then the shape of the output is (N, M, T, 2).
    """
    if kwargs:
        logger.warning(f"Unknown arguments {kwargs!r} - ignoring")
    logger.info("Get displacements using Dual TV-L 1")

    all_flows = []
    for im in np.rollaxis(frames, 2):
        all_flows.append(
            dask.delayed(flow)(im, reference_image, tau, lmbda, theta, nscales, warps),
        )

    with ProgressBar():
        arr = da.compute(all_flows)
    flows = np.stack(*arr, axis=2)  # type:ignore

    logger.info("Done running Dual TV-L 1 algorithm")

    return da.from_array(flows)



import logging
from pathlib import Path
from typing import Optional
from typing import Union

import cv2
import dask.array as da
import matplotlib.pyplot as plt
import numpy as np
import tqdm

from . import frame_sequence as fs
from . import scaling
from . import utils

logger = logging.getLogger(__name__)


def apply_custom_colormap(image_gray, cmap=plt.get_cmap("seismic")):

    assert image_gray.dtype == np.uint8, "must be np.uint8 image"
    if image_gray.ndim == 3:
        image_gray = image_gray.squeeze(-1)

    # Initialize the matplotlib color map
    sm = plt.cm.ScalarMappable(cmap=cmap)

    # Obtain linear color range
    color_range = sm.to_rgba(np.linspace(0, 1, 256))[:, 0:3]  # color range RGBA => RGB
    color_range = (color_range * 255.0).astype(np.uint8)  # [0,1] => [0,255]
    color_range = np.squeeze(
        np.dstack([color_range[:, 2], color_range[:, 1], color_range[:, 0]]),
        0,
    )  # RGB => BGR

    # Apply colormap for each channel individually
    channels = [cv2.LUT(image_gray, color_range[:, i]) for i in range(3)]
    return np.dstack(channels)


def colorize(img, vmin=None, vmax=None, factor=1, cmap="bwr"):
    img = img.astype(float)
    vmin = np.min(img) if vmin is None else vmin
    vmax = np.max(img) if vmax is None else vmax
    img = (img - vmin) / (vmax - vmin)
    img = (img * 255).astype(np.uint8)
    img = apply_custom_colormap(img, cmap=plt.get_cmap(cmap))

    # img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    if factor != 1:
        x, y, _ = img.shape
        img = cv2.resize(img, (y * factor, x * factor))
    return img


def _draw_flow(image, x, y, fx, fy, thickness: int = 2):
    QUIVER = (0, 0, 255)
    lines = np.vstack([x, y, x + fx, y + fy]).T.reshape(-1, 2, 2)
    lines = np.int32(lines + 0.5)
    vis = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
    cv2.polylines(vis, lines, 0, QUIVER, thickness=thickness, lineType=8)
    # for (x1, y1), (_x2, _y2) in lines:
    #     cv2.circle(vis, (x1, y1), 1, (0, 255, 0), -1)
    return vis


def draw_lk_flow(image, flow, reference_points):
    x, y = reference_points.reshape(-1, 2).astype(int).T
    fx, fy = flow.T
    return _draw_flow(image, x, y, fx, fy)


def draw_flow(image, flow, step=16, scale: float = 1.0, thickness: int = 2):
    """[summary]

    Parameters
    ----------
    image : [type]
        [description]
    flow : [type]
        [description]
    step : int, optional
        [description], by default 16

    Returns
    -------
    [type]
        [description]
    """
    h, w = image.shape[:2]
    y, x = np.mgrid[step / 2 : h : step, step / 2 : w : step].reshape(2, -1).astype(int)
    fx, fy = flow[y, x].T
    return _draw_flow(image, x, y, scale * fx, scale * fy, thickness=thickness)


def draw_hsv(flow):
    h, w = flow.shape[:2]
    magnitude, angle = cv2.cartToPolar(flow[..., 0], flow[..., 1])
    hsv = np.zeros((h, w, 3), np.uint8)
    # Sets image hue according to the optical flow
    # direction
    hsv[..., 0] = angle * 180 / np.pi / 2

    # Sets image saturation to maximum
    hsv[..., 1] = 255

    # Sets image value according to the optical flow
    # magnitude (normalized)
    hsv[..., 2] = cv2.normalize(magnitude, None, 0, 255, cv2.NORM_MINMAX)

    # Converts HSV to RGB (BGR) color representation
    rgb = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)

    return rgb


def quiver_video(
    data: utils.MPSData,
    vectors: Union[np.ndarray, fs.VectorFrameSequence],
    path: utils.PathLike,
    step: int = 16,
    vector_scale: float = 1.0,
    frame_scale: float = 1.0,
    convert: bool = True,
    offset: int = 0,
    thickness: int = 2,
) -> None:

    fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    if isinstance(vectors, fs.VectorFrameSequence):
        vectors_np: np.ndarray = vectors.array_np
    elif isinstance(vectors, da.Array):
        vectors_np = vectors.compute()
    else:
        vectors_np = vectors

    assert len(vectors_np.shape) == 4

    if frame_scale < 1.0:
        data = scaling.resize_data(data, frame_scale)

    width = data.size_y
    height = data.size_x
    fps = data.framerate

    num_frames = data.num_frames - offset

    # Check shapes
    try:
        time_axis = vectors_np.shape.index(num_frames)
    except ValueError as ex:
        msg = (
            "Time axis for frames and vetors does not match. "
            f"Got frames of shape {data.frames.shape}, and "
            f"vector of shape {vectors_np.shape}."
        )
        raise ValueError(msg) from ex

    vector_shape = (
        vectors_np.shape[0],
        vectors_np.shape[1],
        vectors_np.shape[time_axis] + offset,
    )

    if not data.frames.shape == vector_shape:
        msg = (
            "Shape of frames and vector does not match. "
            f"Got frames of shape {data.frames.shape}, and "
            f"vector of shape {vectors_np.shape}."
            f"Expected frames to have shape {vector_shape}."
        )
        raise ValueError(msg)

    p = Path(path).with_suffix(".mp4")
    if p.is_file():
        p.unlink()
    frames = utils.to_uint8(data.frames)
    out = cv2.VideoWriter(p.as_posix(), fourcc, fps, (width, height))
    for i in tqdm.tqdm(range(num_frames), desc=f"Create quiver video at {p}"):
        im = frames[:, :, i]
        flow = np.take(vectors_np, i, axis=time_axis)
        out.write(
            draw_flow(im, flow, step=step, scale=vector_scale, thickness=thickness),
        )

    out.release()
    cv2.destroyAllWindows()

    if convert:
        import imageio

        tmp_path = p.parent.joinpath(p.stem + "_tmp").with_suffix(".mp4")
        p.rename(tmp_path)
        video = imageio.read(tmp_path)
        imageio.mimwrite(p, [np.swapaxes(d, 0, 1) for d in video.iter_data()], fps=fps)
        tmp_path.unlink()


def hsv_video(
    path: utils.PathLike,
    vectors: Union[np.ndarray, fs.VectorFrameSequence],
    fps: int = 50,
    axis: int = 2,
    convert: bool = False,
) -> None:

    fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    if isinstance(vectors, fs.VectorFrameSequence):
        vectors_np: np.ndarray = vectors.array_np
    else:
        vectors_np = vectors

    assert len(vectors_np.shape) == 4

    width = vectors_np.shape[1]
    height = vectors_np.shape[0]
    assert len(vectors_np.shape) == 3
    num_frames = vectors_np.shape[axis]

    p = Path(path).with_suffix(".mp4")

    out = cv2.VideoWriter(p.as_posix(), fourcc, fps, (width, height))

    for i in tqdm.tqdm(range(num_frames), desc=f"Create HSV movie at {p}"):
        flow = np.take(vectors_np, i, axis=axis)
        out.write(draw_hsv(flow))

    out.release()
    cv2.destroyAllWindows()

    if convert:
        convert_imageio(p, fps)


def convert_imageio(path, fps):
    logger.info("Convert video using imageio")
    import imageio

    tmp_path = path.parent.joinpath(path.stem + "_tmp").with_suffix(".mp4")
    path.rename(tmp_path)
    video = imageio.read(tmp_path)
    imageio.mimwrite(path, video.iter_data(), fps=fps)


def heatmap(
    path: utils.PathLike,
    data: Union[np.ndarray, fs.FrameSequence],
    fps: int = 50,
    axis: int = 2,
    convert: bool = False,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    cmap="bwr",
    transpose: bool = False,
):

    logger.info("Create heatmap video")
    if isinstance(data, fs.FrameSequence):
        data_np: np.ndarray = data.array_np
    else:
        data_np = data

    vmin = vmin if vmin is not None else data_np.min()
    vmax = vmax if vmax is not None else data_np.max()

    assert len(data_np.shape) == 3
    num_frames = data_np.shape[axis]

    fourcc = cv2.VideoWriter_fourcc(*"mp4v")

    width = data_np.shape[1]
    height = data_np.shape[0]
    if transpose:
        width = data_np.shape[0]
        height = data_np.shape[1]

    p = Path(path).with_suffix(".mp4")

    out = cv2.VideoWriter(p.as_posix(), fourcc, fps, (width, height))
    for i in tqdm.tqdm(range(num_frames), desc=f"Create heatmap movie at {p}"):
        heat = np.take(data_np, i, axis=axis)
        if transpose:
            heat = heat.T
        out.write(colorize(heat, vmin=vmin, vmax=vmax, cmap=cmap))

    out.release()
    cv2.destroyAllWindows()
    logger.info("Done creating heatmap video")
    if convert:
        convert_imageio(p, fps)



import numpy as np


def to_image(A):
    # If all are equal
    B = np.array(A).astype(float)
    if ~np.any(B - B.max()):
        return B.astype(np.uint16)
    return (B / B.max()).astype(np.uint16) * (2**16 - 1)


def add_noise(A, noise_level):

    return to_image(A) + to_image(noise_level * np.random.rand(*np.shape(A)))


def create_circle_data(
    x_start=40,
    dx=5,
    y_start=40,
    dy=5,
    r=10,
    line_x=None,
    line_y=None,
    Nx=100,
    Ny=100,
):

    if line_x is None:
        line_x = np.sin(np.linspace(0, np.pi, 100))
    if line_y is None:
        line_y = np.sin(np.linspace(0, np.pi, 100))

    X = x_start + np.multiply(dx, line_x)
    Y = y_start + np.multiply(dy, line_y)

    A = []
    for x, y in zip(X, Y):
        a = np.fromfunction(
            lambda i, j: np.sqrt((i - x) ** 2 + (j - y) ** 2) < r,
            (Nx, Ny),
        )
        A.append(a)

    return to_image(A).T



import numpy as np
from mps_motion import stats


def test_analysis_from_mechanics(mech_trace_obj):

    analysis = stats.analysis_from_mechanics(mech_trace_obj)

    mean_values = analysis.mean
    std = analysis.std
    assert np.isclose(mean_values["max_contraction_velocity"], 5125.387931222401)
    assert np.isclose(mean_values["max_relaxation_velocity"], 4640.270308088497)
    assert np.isclose(
        mean_values["time_between_contraction_and_relaxation"],
        266.8879394531132,
    )
    assert np.isclose(mean_values["u_peaks"], 0.4589306341013011)
    assert np.isclose(mean_values["u_width50"], 260.4451501720163)

    assert np.isclose(std["max_contraction_velocity"], 161.86832160836673)
    assert np.isclose(std["max_relaxation_velocity"], 178.25122955064532)
    assert np.isclose(std["time_between_contraction_and_relaxation"], 4.707889720282436)
    assert np.isclose(std["u_peaks"], 0.0038838100228260463)
    assert np.isclose(std["u_width50"], 1.1983554291640317)


def test_analysis_from_arrays(mech_trace_obj):

    u = mech_trace_obj.u.norm().mean()
    t = mech_trace_obj.t
    v = mech_trace_obj.velocity().norm().mean()

    analysis = stats.analysis_from_arrays(u, v, t)

    mean_values = analysis.mean
    std = analysis.std
    assert np.isclose(mean_values["max_contraction_velocity"], 5125.387931222401)
    assert np.isclose(mean_values["max_relaxation_velocity"], 4640.270308088497)
    assert np.isclose(
        mean_values["time_between_contraction_and_relaxation"],
        266.8879394531132,
    )
    assert np.isclose(mean_values["u_peaks"], 0.4589306341013011)
    assert np.isclose(mean_values["u_width50"], 260.4451501720163)

    assert np.isclose(std["max_contraction_velocity"], 161.86832160836673)
    assert np.isclose(std["max_relaxation_velocity"], 178.25122955064532)
    assert np.isclose(std["time_between_contraction_and_relaxation"], 4.707889720282436)
    assert np.isclose(std["u_peaks"], 0.0038838100228260463)
    assert np.isclose(std["u_width50"], 1.1983554291640317)



from pathlib import Path
from textwrap import dedent
from typing import Optional

import typer

from .cli import main as _main
from .motion_tracking import FLOW_ALGORITHMS

app = typer.Typer(help="Estimate motion in stack of images")


def version_callback(show_version: bool):
    """Prints version information."""
    if show_version:
        from . import __version__, __program_name__

        typer.echo(f"{__program_name__} {__version__}")
        raise typer.Exit()


def license_callback(show_license: bool):
    """Prints license information."""
    if show_license:
        from . import __license__

        typer.echo(f"{__license__}")
        raise typer.Exit()


@app.command(
    help="Run motion analysis on a single file and output results in a directory",
)
def analyze(
    filename: str = typer.Argument(
        ...,
        help=dedent(
            """
        Path to file to be analyzed, typically an .nd2 or .
        czi Brightfield file
        """,
        ),
    ),
    algorithm: FLOW_ALGORITHMS = typer.Option(
        FLOW_ALGORITHMS.farneback,
        help="The algorithm used to estimate motion",
    ),
    reference_frame: str = typer.Option(
        "0",
        "--reference-frame",
        "-rf",
        help=dedent(
            """
        Which frame should be the reference frame when computing the
        displacements. This can either be a number indicating the
        timepoint, or the value 'mean', 'median', 'max' or 'mean'.
        Default: '0' (i.e the first frame)
        """,
        ),
    ),
    scale: float = typer.Option(
        1.0,
        help=dedent(
            """
        Rescale data before running motion track. This is useful if the spatial resoltion
        of the images are large. Scale = 1.0 will keep the original size
        """,
        ),
    ),
    apply_filter: bool = typer.Option(
        True,
        help=dedent(
            """
            If True, set pixels with max displacement lower than the mean maximum displacement
            to zero. This will prevent non-tissue pixels to be included, which is especially
            error prone for velocity estimations, by default True.""",
        ),
    ),
    spacing: int = typer.Option(
        5,
        help=dedent(
            """Spacing between frames in velocity computations, by default 5.
            """,
        ),
    ),
    compute_xy_components: bool = typer.Option(
        False,
        "--xy",
        "-xy",
        help=dedent(
            """
            If True the compute x- and y components of the displacement and
            velocity and plot them as well, by default False.""",
        ),
    ),
    make_displacement_video: bool = typer.Option(
        False,
        "--video-disp",
        help=dedent(
            """
            If True, create video of displacement vectors, by default False.""",
        ),
    ),
    make_velocity_video: bool = typer.Option(
        False,
        "--video-vel",
        help=dedent(
            """
            If True, create video of velocity vectors, by default False.""",
        ),
    ),
    outdir: Optional[str] = typer.Option(
        None,
        "--outdir",
        "-o",
        help=dedent(
            """
        Directory where to store the results. If not provided, a folder with the the same
        as the filename will be created and results will be stored in a subfolder inside
        that called `motion`
        """,
        ),
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="More verbose"),
):
    _main(
        filename=filename,
        algorithm=algorithm,
        reference_frame=reference_frame,
        outdir=outdir,
        scale=scale,
        apply_filter=apply_filter,
        spacing=spacing,
        compute_xy_components=compute_xy_components,
        make_displacement_video=make_displacement_video,
        make_velocity_video=make_velocity_video,
        verbose=verbose,
    )


@app.command(help="Start gui and analyze files in the provided folder")
def gui(
    path: Path = typer.Argument(
        ...,
        exists=True,
        file_okay=False,
        dir_okay=True,
        readable=True,
        resolve_path=True,
        help=dedent(
            """
        Path to folder where the recordings are located
        """,
        ),
    ),
):

    # Make sure we can import the required packages
    from . import gui  # noqa: F401

    gui_path = Path(__file__).parent.joinpath("gui.py")
    import subprocess as sp

    sp.run(["streamlit", "run", gui_path.as_posix(), "--", path.as_posix()])


if __name__ == "__main__":
    raise SystemExit(app())



from itertools import product
from pathlib import Path

import dask.array as da
import numpy as np
import pytest

try:
    import mps  # noqa: F401

    MPS_NOT_FOUND = False
except ImportError:
    MPS_NOT_FOUND = True

from mps_motion import frame_sequence as fs
from mps_motion import filters

array_type = {da: da.core.Array, np: np.ndarray}


@pytest.mark.parametrize("ns", [np, da])
def test_frame_sequence(ns):
    width = 10
    height = 12
    num_time_steps = 14

    x = fs.FrameSequence(ns.random.random((width, height, num_time_steps)))

    x_mean = x.mean()
    assert isinstance(x_mean, array_type[ns])
    assert x.mean().shape == (num_time_steps,)

    x_max = x.max()
    assert isinstance(x_max, array_type[ns])
    assert x_max.shape == (width, height)


@pytest.mark.parametrize("ns", [np, da])
def test_vector_frame_sequence(ns):
    width = 10
    height = 12
    num_time_steps = 14

    arr = ns.random.random((width, height, num_time_steps, 2))

    x = fs.VectorFrameSequence(arr)

    x_mean = x.mean()
    assert isinstance(x_mean, array_type[ns])
    assert x.mean().shape == (num_time_steps, 2)

    x_max = x.max()
    assert isinstance(x_max, array_type[ns])
    assert x_max.shape == (width, height, 2)

    x_norm = x.norm()
    assert isinstance(x_norm, fs.FrameSequence)
    assert x_norm.shape == (width, height, num_time_steps)

    x_x = x.x
    assert isinstance(x_x, fs.FrameSequence)
    assert x_x.shape == (width, height, num_time_steps)
    assert x_x == fs.FrameSequence(arr[:, :, :, 0])

    x_y = x.y
    assert isinstance(x_y, fs.FrameSequence)
    assert x_y.shape == (width, height, num_time_steps)
    assert x_y == fs.FrameSequence(arr[:, :, :, 1])


@pytest.mark.parametrize("ns", [np, da])
def test_tensor_frame_sequence(ns):

    width = 10
    height = 12
    num_time_steps = 14

    arr = ns.random.random((width, height, num_time_steps, 2, 2))

    x = fs.TensorFrameSequence(arr)

    x_x = x.x
    assert isinstance(x_x, fs.FrameSequence)
    assert x_x.shape == (width, height, num_time_steps)
    assert x_x == fs.FrameSequence(arr[:, :, :, 1, 1])

    x_y = x.y
    assert isinstance(x_y, fs.FrameSequence)
    assert x_y.shape == (width, height, num_time_steps)
    assert x_y == fs.FrameSequence(arr[:, :, :, 0, 0])

    x_xy = x.xy
    assert isinstance(x_xy, fs.FrameSequence)
    assert x_xy.shape == (width, height, num_time_steps)
    assert x_xy == fs.FrameSequence(arr[:, :, :, 1, 0])

    x_norm = x.norm()
    assert isinstance(x_norm, fs.FrameSequence)
    assert x_norm.shape == (width, height, num_time_steps)

    x_mean = x.mean()
    assert isinstance(x_mean, array_type[ns])
    assert x.mean().shape == (num_time_steps, 2, 2)

    x_max = x.max()
    assert isinstance(x_max, array_type[ns])
    assert x_max.shape == (width, height, 2, 2)

    try:
        # If we have a dask array
        a = arr.compute()
    except AttributeError:
        a = arr

    q = np.linalg.eigvalsh(a)
    p = x.compute_eigenvalues()
    assert p.shape == q.shape
    assert np.isclose(p.array, q).all()


@pytest.mark.skipif(MPS_NOT_FOUND, reason="MPS not found")
@pytest.mark.parametrize("ns", [np, da])
def test_local_averages(ns):

    width = 10
    height = 15
    num_time_steps = 14

    arr = ns.ones((width, height, num_time_steps))

    x = fs.FrameSequence(arr)

    N = 3

    la = x.local_averages(N)
    assert la.shape == (width // (height // N), N, num_time_steps)


@pytest.mark.parametrize("ns, suffix", product([np, da], [".h5", ".npy"]))
def test_save_load(ns, suffix):

    width = 10
    height = 15
    num_time_steps = 14

    arr = ns.ones((width, height, num_time_steps))

    x = fs.FrameSequence(arr)
    path = Path("test").with_suffix(suffix)
    x.save(path)

    new_x = fs.FrameSequence.from_file(path)

    assert x == new_x
    path.unlink()


@pytest.mark.parametrize("ns", [np, da])
def test_norm(ns):

    width = 10
    height = 15
    num_time_steps = 14

    arr = ns.ones((width, height, num_time_steps, 2))

    x = fs.VectorFrameSequence(arr)
    x_norm = x.norm()
    assert x_norm.shape == (width, height, num_time_steps)
    assert isinstance(x_norm, fs.FrameSequence)
    arr = x_norm.compute()
    assert (arr == np.sqrt(2)).all()
    # breakpoint()


@pytest.mark.parametrize(
    "ns, limits",
    product([np, da], [(0.4, 0.6), (None, 0.6), (0.4, None), (0.5, 0.5), (None, None)]),
)
def test_threshold(ns, limits):
    vmin, vmax = limits

    width = 10
    height = 15
    num_time_steps = 14

    np.random.seed(1)
    values = ns.random.random((width, height, num_time_steps))

    # This value should remain the same
    special_value = 0.5
    # Norm of this is about 0.56
    values[0, 0, 0] = special_value

    arr = fs.FrameSequence(values)
    th_arr = arr.threshold(vmin, vmax)
    if vmin is None:
        vmin = values.min()
    if vmax is None:
        vmax = values.max()

    assert ns.isclose(th_arr.max().max(), vmax)
    assert ns.isclose(th_arr.min().min(), vmin)
    assert ns.isclose(th_arr.array[0, 0, 0], special_value)


@pytest.mark.parametrize("ns", [np, da])
def test_invalid_threshold_raises(ns):

    width = 10
    height = 15
    num_time_steps = 14

    values = ns.random.random((width, height, num_time_steps))
    arr = fs.FrameSequence(values)

    with pytest.raises(filters.InvalidThresholdError):
        arr.threshold(0.7, 0.2)


@pytest.mark.parametrize(
    "ns, limits",
    product([np, da], [(0.4, 0.6), (None, 0.6), (0.4, None), (0.5, 0.5), (None, None)]),
)
def test_threshold_norm(ns, limits):
    vmin, vmax = limits

    width = 10
    height = 15
    num_time_steps = 14

    np.random.seed(1)
    values = ns.random.random((width, height, num_time_steps, 2))

    # This value should remain the same
    special_value = np.sqrt(0.125)
    # sqrt(0.125 + 0.125) = sqrt(0.25) = 0.5
    values[0, 0, 0, 0] = special_value
    values[0, 0, 0, 1] = special_value

    arr = fs.VectorFrameSequence(values)
    th_arr = arr.threshold_norm(vmin, vmax)
    if vmin is None:
        vmin = arr.norm().min().min()
    if vmax is None:
        vmax = arr.norm().max().max()

    assert ns.isclose(th_arr.norm().max().max(), vmax)
    assert ns.isclose(th_arr.norm().min().min(), vmin)
    assert ns.isclose(th_arr.array[0, 0, 0, 0], special_value)
    assert ns.isclose(th_arr.array[0, 0, 0, 1], special_value)


@pytest.mark.parametrize(
    "filter_type, size, sigma",
    [
        (filters.Filters.median, 3, None),
        (filters.Filters.gaussian, None, 1),
    ],
)
def test_filter_VectorFrameSequence(filter_type, size, sigma):

    shape = (10, 9, 8, 2)
    np.random.seed(1)
    vectors = 10 * np.ones(shape) + np.random.random(shape)

    u = fs.VectorFrameSequence(vectors)

    u_filt = u.filter(filter_type=filter_type, size=size, sigma=sigma)
    filtered_vectors = u_filt.array

    assert filtered_vectors.shape == shape

    assert 0 < np.abs(filtered_vectors - vectors).max() < 1


@pytest.mark.parametrize(
    "filter_type, size, sigma",
    [
        (filters.Filters.median, 3, None),
        (filters.Filters.gaussian, None, 1),
    ],
)
def test_filter_FrameSequence(filter_type, size, sigma):

    shape = (10, 9, 8)
    np.random.seed(1)
    vectors = 10 * np.ones(shape) + np.random.random(shape)

    u = fs.FrameSequence(vectors)

    u_filt = u.filter(filter_type=filter_type, size=size, sigma=sigma)
    filtered_vectors = u_filt.array

    assert filtered_vectors.shape == shape

    assert 0 < np.abs(filtered_vectors - vectors).max() < 1


def test_cartToPolar():

    # a = 0.01
    # b = 0.01
    # def func(x):
    #     return (x[0] * (1 - a), x[1] * (1 - b))

    # frame1 = np.random.randint(size=(12, 12))
    # from scipy.ndimage import geometric_transform
    # frame2 = geometric_transform(frame1.T, func).T
    # np.fromfunction
    pass


if __name__ == "__main__":
    test_threshold_norm(da, (0.4, 0.6))



import numpy as np
from mps_motion import Mechanics
from mps_motion import OpticalFlow
from mps_motion import scaling
from mps_motion import utils


def test_resize_data():
    frames = np.ones((100, 100, 3))
    times = np.linspace(0, 5000, frames.shape[-1])
    info = dict(um_per_pixel=1.0)
    data = utils.MPSData(frames=frames, time_stamps=times, info=info)
    scale = 0.5
    scaled_data = scaling.resize_data(data, scale)
    assert scaled_data.info["um_per_pixel"] == info["um_per_pixel"] / scale
    assert scaled_data.frames.shape == (50, 50, 3)
    assert np.isclose(scaled_data.time_stamps, times).all()


def test_resize_frames_displacement():
    width = 10
    height = 15
    num_time_points = 4
    scale = 0.4

    np.random.seed(1)
    u = np.zeros((height, width, 2, num_time_points))
    u[:, :, 0, 0] = np.fromfunction(
        lambda y, x: x / width,
        shape=(height, width),
        dtype=float,
    )
    u[:, :, 1, 0] = np.fromfunction(
        lambda y, x: y / height,
        shape=(height, width),
        dtype=float,
    )
    u_x_resized = scaling.resize_frames(u[:, :, 0, :], scale=scale)
    u_y_resized = scaling.resize_frames(u[:, :, 1, :], scale=scale)
    u_resized = np.stack((u_x_resized, u_y_resized), axis=2)

    u_x_orig = scaling.resize_frames(u_x_resized, scale=1 / scale)
    u_y_orig = scaling.resize_frames(u_y_resized, scale=1 / scale)
    u_orig = np.stack((u_x_orig, u_y_orig), axis=2)

    assert u_orig.shape == u.shape
    assert abs(u[:, :, 0, 0].mean() - u_resized[:, :, 0, 0].mean()) < 1e-12
    assert abs(u[:, :, 1, 0].mean() - u_resized[:, :, 1, 0].mean()) < 1e-12

    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots(1, 4)
    # ax[0].imshow(u[:, :, 1, 0])
    # ax[1].imshow(u_resized[:, :, 1, 0])
    # ax[2].imshow(u_orig[:, :, 1, 0])
    # ax[3].imshow((u - u_orig)[:, :, 1, 0])
    # plt.show()


def _test_resize_frames_units():
    """This is a visual test"""
    import matplotlib.pyplot as plt
    import synthetic_data

    frames = synthetic_data.create_circle_data(dx=5, dy=0, Nx=200, Ny=200)
    times = np.linspace(0, 5000, frames.shape[-1])
    info = dict(um_per_pixel=1.0)

    # mps.utils.frames2mp4(frames.T, "synthetic_data")
    data = utils.MPSData(frames=frames, time_stamps=times, info=info)
    # new_data = scaling.resize_data(data, scale=0.5)

    # mps.utils.frames2mp4(scaling.resize_frames(frames, 0.5).T, "synthetic_data_05")
    flow = OpticalFlow(data, "farneback")

    # scale = 1.0
    for scale in [0.5, 0.8, 1.0]:
        flow = OpticalFlow(data, "farneback")
        u = flow.get_displacements(scale=scale)  # , unit="pixels")

        mech = Mechanics(u)

        # import dask.array as da
        U = mech.E.compute_eigenvalues()

        # U = mech.u

        u_norm = U.norm().array  # .compute()
        u_norm_time = U.norm().mean()  # .compute()

        # from IPython import embed

        # embed()
        # exit()

        # u_norm = mech.E.x.array.compute()
        # u_norm_time = mech.E.x.mean().compute()

        fig, ax = plt.subplots()
        ax.plot(times, u_norm_time)
        # continue
        # plt.show()

        from mps.analysis import local_averages

        la = local_averages(
            u_norm,
            data.time_stamps,
            background_correction=False,
            N=int(10 * scale),
        )

        fig, ax = plt.subplots(
            la.shape[0],
            la.shape[1],
            sharex=True,
            sharey=True,
            figsize=(12, 12),
        )
        for i in range(la.shape[0]):
            for j in range(la.shape[1]):
                ax[i, j].plot(data.time_stamps, la[i, j, :])

        # mps.utils.frames2mp4(u_norm.T, "disp_norm")

        u_max = np.max(u_norm, axis=2)
        fig, ax = plt.subplots()
        im = ax.imshow(u_max)
        fig.colorbar(im)
    plt.show()


if __name__ == "__main__":
    # test_ScaledDisplacemet()
    # test_resize_frames_units()
    # pass
    _test_resize_frames_units()



import logging
from typing import Optional
from typing import Tuple

import cv2
import dask
import dask.array as da
import numpy as np
import scipy.spatial
import tqdm
from dask.diagnostics import ProgressBar

from .utils import Array
from .utils import MPSData

logger = logging.getLogger(__name__)

INTERPOLATION_METHODS = {
    "nearest": cv2.INTER_NEAREST,
    "linear": cv2.INTER_LINEAR,
    "area": cv2.INTER_AREA,
    "cubic": cv2.INTER_CUBIC,
    "lanczos": cv2.INTER_LANCZOS4,
}


def resize_vectors(vectors, new_shape):
    if len(vectors.shape) == 4:
        vec1 = resize_frames(
            vectors[:, :, :, 0],
            new_shape=new_shape,
        )
        vec2 = resize_frames(
            vectors[:, :, :, 1],
            new_shape=new_shape,
        )
    else:
        assert len(vectors.shape) == 3
        vec1 = resize_frames(
            vectors[:, :, 0],
            new_shape=new_shape,
        )
        vec2 = resize_frames(
            vectors[:, :, 1],
            new_shape=new_shape,
        )
    return da.stack([vec1, vec2], axis=-1)


def resize_data(data: MPSData, scale: float) -> MPSData:
    new_frames = resize_frames(data.frames, scale)
    info = data.info.copy()
    info["um_per_pixel"] /= scale
    info["size_x"], info["size_y"], info["num_frames"] = new_frames.shape
    return MPSData(new_frames, data.time_stamps, info)


def subsample_time(data: MPSData, step: int) -> MPSData:

    new_frames = data.frames[:, :, ::step]
    new_times = data.time_stamps[::step]
    info = data.info.copy()
    info["num_frames"] = len(new_times)

    return MPSData(new_frames, new_times, info)


def reshape_lk(reference_points: np.ndarray, flows: Array) -> Array:
    x, y = reference_points.reshape(-1, 2).astype(int).T
    xu = np.sort(np.unique(x))
    yu = np.sort(np.unique(y))

    is_dask = False
    if isinstance(flows, da.Array):
        is_dask = True

    dx = xu[0]
    dxs = np.diff(xu)
    if len(dxs) > 0:
        dx = dxs[0]
        assert np.all(dxs == dx)

    dy = yu[0]
    dys = np.diff(yu)
    if len(dys) > 0:
        dy = dys[0]
        assert np.all(dys == dy)

    xp = ((x - np.min(x)) / dx).astype(int)
    yp = ((y - np.min(y)) / dy).astype(int)
    if len(flows.shape) == 3:
        num_frames = flows.shape[-1]
        out = np.zeros((yp.max() + 1, xp.max() + 1, 2, num_frames))
        out[yp, xp, :, :] = flows
        out = np.swapaxes(out, 2, 3)
    else:
        out = np.zeros((yp.max() + 1, xp.max() + 1, 2))
        out[yp, xp, :] = flows

    if is_dask:
        out = da.from_array(out)
    return out


def resize_frames(
    frames: np.ndarray,
    scale: float = 1.0,
    new_shape: Optional[Tuple[int, int]] = None,
    interpolation_method="nearest",
) -> np.ndarray:
    logger.info("Resize frames")
    msg = f"Expected interpolation method to be one of {INTERPOLATION_METHODS.keys()}, got {interpolation_method}"
    assert interpolation_method in INTERPOLATION_METHODS, msg
    if scale != 1.0 or new_shape is not None:

        if len(frames.shape) == 2:
            w, h = frames.shape
        else:
            w, h, num_frames = frames.shape

        if new_shape is not None:
            assert len(new_shape) == 2
            width, height = new_shape
        else:
            width = int(w * scale)
            height = int(h * scale)

        width = int(width)
        height = int(height)

        if len(frames.shape) == 2:
            return cv2.resize(frames, (height, width))

        all_resized_frames = []
        for i in range(num_frames):
            all_resized_frames.append(
                dask.delayed(cv2.resize)(
                    frames[:, :, i],
                    (height, width),
                    INTERPOLATION_METHODS[interpolation_method],
                ),
            )
        with ProgressBar():
            resized_frames = da.stack(
                *da.compute(all_resized_frames), axis=-1
            ).compute()

    else:
        resized_frames = frames.copy()
    logger.info("Done resizing")
    return resized_frames


def interpolate_lk_flow(
    disp: np.ndarray,
    reference_points: np.ndarray,
    size_x: int,
    size_y: int,
    interpolation_method: str = "linear",
) -> np.ndarray:
    """Given an array of displacements (of flow) coming from
    the Lucas Kanade method return a new array which
    interpolates the data onto a given size, i.e
    the original size of the image.

    Parameters
    ----------
    disp : np.ndarray
        The flow or displacement from LK algorithm
    reference_points : np.ndarray
        Reference points
    size_x : int
        Size of the output in x-direction
    size_y : int
        Size of the output in y-direction
    interpolation_method : str
        Method for interpolation, by default 'linear'

    Returns
    -------
    np.ndarray
        Interpolated values
    """
    num_frames = disp.shape[-1]
    from scipy.interpolate import griddata

    disp_full = np.zeros((size_y, size_x, 2, num_frames))
    ref_points = np.squeeze(reference_points)
    grid_x, grid_y = np.meshgrid(np.arange(size_x), np.arange(size_y))
    # TODO: This could be parallelized
    for i in tqdm.tqdm(range(num_frames)):
        values_x = disp[:, 0, i]
        values_y = disp[:, 1, i]

        disp_full[:, :, 0, i] = griddata(
            ref_points,
            values_x,
            (grid_y, grid_x),
            method=interpolation_method,
        )
        disp_full[:, :, 1, i] = griddata(
            ref_points,
            values_y,
            (grid_y, grid_x),
            method=interpolation_method,
        )
    return disp_full


def rbfinterp2d_map(args):
    return rbfinterp2d(*args)


def rbfinterp2d(  # noqa:C901
    coord,
    input_array,
    xgrid,
    ygrid,
    rbfunction="gaussian",
    epsilon=10,
    k=50,
    nchunks=5,
):
    """
    Fast 2-D grid interpolation of a sparse (multivariate) array using a
    radial basis function.

    https://docs.scipy.org/doc/numpy/reference/generated/numpy.ndarray.html

    Parameters
    ----------
    coord: array_like
        Array of shape (n, 2) containing the coordinates of the data points
        into a 2-dimensional space.
    input_array: array_like
        Array of shape (n) or (n, m) containing the values of the data points,
        where *n* is the number of data points and *m* the number of co-located
        variables. All values in ``input_array`` are required to have finite values.
    xgrid, ygrid: array_like
        1D arrays representing the coordinates of the 2-D output grid.
    rbfunction: {"gaussian", "multiquadric", "inverse quadratic", "inverse multiquadric", "bump"}, optional
        The name of one of the available radial basis function based on a
        normalized Euclidian norm as defined in the **Notes** section below.
        More details provided in the wikipedia reference page linked below.
    epsilon: float, optional
        The shape parameter used to scale the input to the radial kernel.
        A smaller value for ``epsilon`` produces a smoother interpolation. More
        details provided in the wikipedia reference page linked below.
    k: int or None, optional
        The number of nearest neighbours used for each target location.
        This can also be useful to to speed-up the interpolation.
        If set to None, it interpolates using all the data points at once.
    nchunks: int, optional
        The number of chunks in which the grid points are split to limit the
        memory usage during the interpolation.

    Returns
    -------
    output_array: ndarray
        The interpolated field(s) having shape (*m*, ``ygrid.size``, ``xgrid.size``).

    Notes
    -----
    The coordinates are normalized before computing the Euclidean norms:

    .. math::
        x = (x - min(x)) / max[max(x) - min(x), max(y) - min(y)],
        y = (y - min(y)) / max[max(x) - min(x), max(y) - min(y)],

    where the min and max values are taken as the 2nd and 98th percentiles.

    References
    ----------
    Wikipedia contributors, "Radial basis function,"
    Wikipedia, The Free Encyclopedia,
    https://en.wikipedia.org/w/index.php?title=Radial_basis_function&oldid=906155047
    (accessed August 19, 2019).

    """

    _rbfunctions = [
        "nearest",
        "gaussian",
        "inverse quadratic",
        "inverse multiquadric",
        "bump",
    ]

    input_array = np.copy(input_array)

    if np.any(~np.isfinite(input_array)):
        raise ValueError("input_array contains non-finite values")

    if input_array.ndim == 1:
        nvar = 1
        input_array = input_array[:, None]

    elif input_array.ndim == 2:
        nvar = input_array.shape[1]

    else:
        raise ValueError(
            "input_array must have 1 (n) or 2 dimensions (n, m), but it has %i"
            % input_array.ndim,
        )

    npoints = input_array.shape[0]

    if npoints == 0:
        raise ValueError(
            "input_array (n, m) must contain at least one sample, but it has %i"
            % npoints,
        )

    # only one sample, return uniform fields
    elif npoints == 1:
        output_array = np.ones((nvar, ygrid.size, xgrid.size))
        for i in range(nvar):
            output_array[i, :, :] *= input_array[:, i]
        return output_array

    coord = np.copy(coord)

    if coord.ndim != 2:
        raise ValueError(
            "coord must have 2 dimensions (n, 2), but it has %i" % coord.ndim,
        )

    if npoints != coord.shape[0]:
        raise ValueError(
            "the number of samples in the input_array does not match the "
            + "number of coordinates %i!=%i" % (npoints, coord.shape[0]),
        )

    # normalize coordinates
    qcoord = np.percentile(coord, [2, 98], axis=0)
    dextent = np.max(np.diff(qcoord, axis=0))
    coord = (coord - qcoord[0, :]) / dextent

    rbfunction = rbfunction.lower()
    if rbfunction not in _rbfunctions:
        raise ValueError(
            "Unknown rbfunction '{}'\n".format(rbfunction)
            + "The available rbfunctions are: "
            + str(_rbfunctions),
        ) from None

    # generate the target grid
    X, Y = np.meshgrid(xgrid, ygrid)
    grid = np.column_stack((X.ravel(), Y.ravel()))
    # normalize the grid coordinates
    grid = (grid - qcoord[0, :]) / dextent

    # k-nearest interpolation
    if k is not None and k > 0:
        k = int(np.min((k, npoints)))

        # create cKDTree object to represent source grid
        tree = scipy.spatial.cKDTree(coord)

    else:
        k = 0

    # split grid points in n chunks
    if nchunks > 1:
        subgrids = np.array_split(grid, nchunks, 0)
        subgrids = [x for x in subgrids if x.size > 0]

    else:
        subgrids = [grid]

    # loop subgrids
    i0 = 0
    output_array = np.zeros((grid.shape[0], nvar))
    for i, subgrid in enumerate(subgrids):
        idelta = subgrid.shape[0]

        if k == 0:
            # use all points
            d = scipy.spatial.distance.cdist(coord, subgrid, "euclidean").transpose()
            inds = np.arange(npoints)[None, :] * np.ones(
                (subgrid.shape[0], npoints),
            ).astype(int)

        else:
            # use k-nearest neighbours
            d, inds = tree.query(subgrid, k=k)

        if k == 1:
            # nearest neighbour
            output_array[i0 : (i0 + idelta), :] = input_array[inds, :]

        else:

            # the interpolation weights
            if rbfunction == "gaussian":
                w = np.exp(-((d * epsilon) ** 2))

            elif rbfunction == "inverse quadratic":
                w = 1.0 / (1 + (epsilon * d) ** 2)

            elif rbfunction == "inverse multiquadric":
                w = 1.0 / np.sqrt(1 + (epsilon * d) ** 2)

            elif rbfunction == "bump":
                w = np.exp(-1.0 / (1 - (epsilon * d) ** 2))
                w[d >= 1 / epsilon] = 0.0

            if not np.all(np.sum(w, axis=1)):
                w[np.sum(w, axis=1) == 0, :] = 1.0

            # interpolate
            for j in range(nvar):
                output_array[i0 : (i0 + idelta), j] = np.sum(
                    w * input_array[inds, j],
                    axis=1,
                ) / np.sum(w, axis=1)

        i0 += idelta

    # reshape to final grid size
    return output_array.reshape(ygrid.size, xgrid.size, nvar)



import dask.array as da
import numpy as np
import pytest
from mps_motion import frame_sequence as fs
from mps_motion import Mechanics
from mps_motion import mechanics


def test_deformation_gradient():

    width = 10
    height = 15

    a11 = 1 / width
    a21 = -1 / height
    a12 = -1 / width
    a22 = 1 / height

    u = np.zeros((width, height, 1, 2))
    u[:, :, 0, 0] = np.fromfunction(
        lambda x, y: a11 * x + a21 * y,
        shape=(width, height),
        dtype=float,
    )
    u[:, :, 0, 1] = np.fromfunction(
        lambda x, y: a12 * x + a22 * y,
        shape=(width, height),
        dtype=float,
    )
    dx = 1
    U = fs.VectorFrameSequence(u, dx=dx)
    m = Mechanics(U)

    du = np.array([[a11 / dx, a12 / dx], [a21 / dx, a22 / dx]])
    F = du + np.eye(2)

    assert da.isclose(m.F[:, :, :, 0, 0], F[0, 0]).all().compute()
    assert da.isclose(m.F[:, :, :, 0, 1], F[0, 1]).all().compute()
    assert da.isclose(m.F[:, :, :, 1, 0], F[1, 0]).all().compute()
    assert da.isclose(m.F[:, :, :, 1, 1], F[1, 1]).all().compute()


def test_strain():

    width = 10
    height = 15

    a11 = 1 / width
    a21 = -1 / height
    a12 = -1 / width
    a22 = 1 / height

    u = np.zeros((width, height, 1, 2))
    u[:, :, 0, 0] = np.fromfunction(
        lambda x, y: a11 * x + a21 * y,
        shape=(width, height),
        dtype=float,
    )
    u[:, :, 0, 1] = np.fromfunction(
        lambda x, y: a12 * x + a22 * y,
        shape=(width, height),
        dtype=float,
    )
    dx = 1
    U = fs.VectorFrameSequence(u, dx=dx)
    m = Mechanics(U)

    du = np.array([[a11 / dx, a12 / dx], [a21 / dx, a22 / dx]])
    F = du + np.eye(2)
    E = 0.5 * (F.T.dot(F) - np.eye(2))

    assert da.isclose(m.E[:, :, :, 0, 0], E[0, 0]).all().compute()
    assert da.isclose(m.E[:, :, :, 0, 1], E[0, 1]).all().compute()
    assert da.isclose(m.E[:, :, :, 1, 0], E[1, 0]).all().compute()
    assert da.isclose(m.E[:, :, :, 1, 1], E[1, 1]).all().compute()


def test_principal_strain(mech_obj):
    e1 = mech_obj.principal_strain[:, :, :, 0]

    assert np.allclose(e1[:, :, 0], 0)

    a1 = np.array([[0.005, 0.05], [0.05, 0.0]])
    eig = np.linalg.eigvals(a1)
    assert np.allclose(e1[:, :, 1], eig.min())

    e2 = mech_obj.principal_strain[:, :, :, 1]

    assert np.allclose(e2[:, :, 1], eig.max())


@pytest.mark.parametrize("dx", [1.0, 0.5, 2.0])
def test_dx(dx):
    # f(x, y) = (x / width - y / height & y / height - x / width)
    # Df = (1/width & -1/height // -1/width & 1/height)

    width = 10
    height = 15

    a11 = 1 / width
    a21 = -1 / height
    a12 = -1 / width
    a22 = 1 / height

    u = np.zeros((width, height, 1, 2))
    u[:, :, 0, 0] = np.fromfunction(
        lambda x, y: a11 * x + a21 * y,
        shape=(width, height),
        dtype=float,
    )
    u[:, :, 0, 1] = np.fromfunction(
        lambda x, y: a12 * x + a22 * y,
        shape=(width, height),
        dtype=float,
    )
    dx = 1
    U = fs.VectorFrameSequence(u, dx=dx)
    m = Mechanics(U)

    assert da.isclose(m.du[:, :, :, 0, 0], a11 / dx).all().compute()
    assert da.isclose(m.du[:, :, :, 0, 1], a12 / dx).all().compute()
    assert da.isclose(m.du[:, :, :, 1, 0], a21 / dx).all().compute()
    assert da.isclose(m.du[:, :, :, 1, 1], a22 / dx).all().compute()


def test_velocity(mech_obj):
    v = mech_obj.velocity() * 1000
    assert np.isclose(
        v[:, :, 0, 0],
        mech_obj.u[:, :, 1, 0] - mech_obj.u[:, :, 0, 0],
    ).all()
    assert np.isclose(
        v[:, :, 1, 0],
        mech_obj.u[:, :, 2, 0] - mech_obj.u[:, :, 1, 0],
    ).all()
    assert np.isclose(
        v[:, :, 0, 1],
        mech_obj.u[:, :, 1, 1] - mech_obj.u[:, :, 0, 1],
    ).all()
    assert np.isclose(
        v[:, :, 1, 1],
        mech_obj.u[:, :, 2, 1] - mech_obj.u[:, :, 1, 1],
    ).all()


def test_compute_displacement(mech_obj):
    spacing = 1  # Only implemented for spacings of 1
    v = mech_obj.velocity(spacing=spacing)
    u = mechanics.compute_displacement(
        v.array,
        mech_obj.t,
        ref_index=0,
        spacing=spacing,
    )
    assert da.isclose(u, mech_obj.u.array).all().compute()


def test_shapes():

    width = 10
    height = 12
    num_time_steps = 14

    u = np.random.random((width, height, num_time_steps, 2))
    m = Mechanics(fs.VectorFrameSequence(u))

    assert m.u.shape == (width, height, num_time_steps, 2)
    assert m.du.shape == (width, height, num_time_steps, 2, 2)
    assert m.F.shape == (width, height, num_time_steps, 2, 2)
    assert m.E.shape == (width, height, num_time_steps, 2, 2)
    assert m.principal_strain.shape == (width, height, num_time_steps, 2)


if __name__ == "__main__":
    # main()
    # mech_obj()
    # test_shapes()
    test_dx(0.5)



import logging
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple

import ap_features as apf
import dask.array as da
import numpy as np
from scipy.signal import find_peaks

from .mechanics import Mechanics
from .utils import Array

logger = logging.getLogger(__name__)


class Analysis(NamedTuple):
    u_peaks: List[float]
    u_width50: List[float]
    time_between_contraction_and_relaxation: List[float]
    max_contraction_velocity: List[float]
    max_relaxation_velocity: List[float]

    @property
    def mean(self) -> Dict[str, float]:
        return {
            "u_peaks": np.mean(self.u_peaks),
            "u_width50": np.mean(self.u_width50),
            "time_between_contraction_and_relaxation": np.mean(
                self.time_between_contraction_and_relaxation,
            ),
            "max_contraction_velocity": np.mean(self.max_contraction_velocity),
            "max_relaxation_velocity": np.mean(self.max_relaxation_velocity),
        }

    @property
    def std(self) -> Dict[str, float]:
        return {
            "u_peaks": np.std(self.u_peaks),
            "u_width50": np.std(self.u_width50),
            "time_between_contraction_and_relaxation": np.std(
                self.time_between_contraction_and_relaxation,
            ),
            "max_contraction_velocity": np.std(self.max_contraction_velocity),
            "max_relaxation_velocity": np.std(self.max_relaxation_velocity),
        }


class ProminenceError(RuntimeError):
    pass


def normalize(y: np.ndarray) -> np.ndarray:
    y = np.array(y)
    return (y - y.min()) / (y.max() - y.min())


def width_at_height(y: np.ndarray, t: np.ndarray, height=0.5) -> float:
    pass
    # assert 0 <= height <= 1
    # f = UnivariateSpline(t, normalize(y) - height, s=0, k=3)
    # try:
    #     return np.diff(f.roots())[0]
    # except IndexError:
    #     # Function does not have two zeros
    #     return np.nan


def chop_trace(
    u: np.ndarray,
    t: np.ndarray,
    pacing: Optional[np.ndarray] = None,
    background_correction_method="subtract",
    zero_index: Optional[int] = None,
    ignore_pacing: bool = False,
    intervals: Optional[List[apf.chopping.Interval]] = None,
) -> Tuple[List[apf.Beat], Optional[List[apf.chopping.Interval]]]:
    trace = apf.Beats(
        u,
        t,
        pacing,
        background_correction_method=background_correction_method,
        zero_index=zero_index,
        chopping_options={"ignore_pacing": ignore_pacing},
        intervals=intervals,
    )

    try:
        beats = trace.beats
        intervals = trace.intervals
    except (apf.chopping.InvalidChoppingError, apf.chopping.EmptyChoppingError):
        beats = [trace.as_beat()]

    return beats, intervals


def compute(u: Array) -> np.ndarray:
    if isinstance(u, da.Array):
        return u.compute()
    return u


def find_two_most_prominent_peaks(y) -> Tuple[int, int]:
    """Return the indices of the two most promintent peaks

    Parameters
    ----------
    y : np.ndarray
        The signals

    Returns
    -------
    Tuple[int, int]
        Indices of the first and second peak
    """
    from scipy.signal import find_peaks

    peaks: List[int] = []
    prominence = 0.9
    while len(peaks) < 2 and prominence > 0:
        peaks, opts = find_peaks(normalize(y), prominence=prominence)
        prominence -= 0.1

    if len(peaks) < 2:
        raise ProminenceError("Unable to find two most prominent beats")

    return (peaks[0], peaks[1])


def analysis_from_mechanics(
    mech: Mechanics,
    pacing: Optional[np.ndarray] = None,
    background_correction_method="subtract",
    zero_index: Optional[int] = None,
    ignore_pacing: bool = False,
    intervals: Optional[List[apf.chopping.Interval]] = None,
) -> Analysis:

    u = compute(mech.u.norm().mean())
    t = mech.t
    v = compute(mech.velocity().norm().mean())
    return analysis_from_arrays(
        u,
        v,
        t,
        pacing=pacing,
        background_correction_method=background_correction_method,
        zero_index=zero_index,
        ignore_pacing=ignore_pacing,
        intervals=intervals,
    )


def analysis_from_arrays(
    u: np.ndarray,
    v: np.ndarray,
    t: np.ndarray,
    pacing: Optional[np.ndarray] = None,
    background_correction_method="subtract",
    zero_index: Optional[int] = None,
    ignore_pacing: bool = False,
    intervals: Optional[List[apf.chopping.Interval]] = None,
) -> Analysis:

    u = apf.utils.numpyfy(u)
    v = apf.utils.numpyfy(v)
    t = apf.utils.numpyfy(t)

    u_beats, intervals = chop_trace(
        u,
        t,
        pacing=pacing,
        intervals=intervals,
        zero_index=zero_index,
        background_correction_method=background_correction_method,
        ignore_pacing=ignore_pacing,
    )

    u_peaks = [ui.y.max() for ui in u_beats]
    # u_ttp = [ui.ttp() for ui in u_beats]
    u_width50 = [ui.apd(50) for ui in u_beats]
    p = pacing
    if pacing is not None:
        p = pacing[: len(v)]

    v_beats, intervals = chop_trace(
        v,
        t[: len(v)],
        pacing=p,
        intervals=intervals,
        zero_index=zero_index,
        background_correction_method=background_correction_method,
        ignore_pacing=ignore_pacing,
    )
    # time_to_max_contraction_velocty = []
    max_contraction_velocity = []
    # time_to_max_relaxation_velocty = []
    max_relaxation_velocity = []
    time_between_contraction_and_relaxation = []

    for vi in v_beats:
        try:
            t0, t1 = find_two_most_prominent_peaks(vi.y)
        except ProminenceError:
            time_between_contraction_and_relaxation.append(0.0)
            # time_to_max_contraction_velocty.append(vi.t[t0])
            max_contraction_velocity.append(0.0)
            # time_to_max_relaxation_velocty.append(vi.t[t1])
            max_relaxation_velocity.append(0.0)

        else:
            time_between_contraction_and_relaxation.append(vi.t[t1] - vi.t[t0])
            # time_to_max_contraction_velocty.append(vi.t[t0])
            max_contraction_velocity.append(vi.y[t0])
            # time_to_max_relaxation_velocty.append(vi.t[t1])
            max_relaxation_velocity.append(vi.y[t1])
    return Analysis(
        u_peaks=u_peaks,
        u_width50=u_width50,
        max_contraction_velocity=max_contraction_velocity,
        max_relaxation_velocity=max_relaxation_velocity,
        time_between_contraction_and_relaxation=time_between_contraction_and_relaxation,
    )


def find_two_peaks(y, prominence=1.0):
    peaks, peak_ops = find_peaks(y, prominence=prominence)

    if len(peaks) == 2:
        return peaks, peak_ops, prominence
    elif len(peaks) < 2:
        # We need to lower the prominence
        while len(peaks) < 2 and prominence > 0:
            prominence -= 0.1
            peaks, peak_ops = find_peaks(y, prominence=prominence)
    else:
        # We need to increase the prominence
        while len(peaks) > 2 and prominence < 10.0:
            prominence += 0.1
            peaks, peak_ops = find_peaks(y, prominence=prominence)

    return peaks, peak_ops, prominence


def compute_features(u, v, t):
    u = apf.Beats(u, t, background_correction_method="subtract")
    u_beats = u.beats
    v = apf.Beats(
        v,
        t[: len(v)],
        intervals=u.chopped_data.intervals,
        background_correction_method="subtract",
    )

    v_beats = v.beats

    data = {
        "Maximum rise velocity": [],
        "Peak twitch amplitude": [],
        "Maximum relaxation velocity": [],
        "Beat duration": [],
        "Time to peak twitch amplitude": [],
        "Time to peak contraction velocity": [],
        "Time to peak relaxation velocity": [],
        "Width at half height": [],
    }

    for ui, vi in zip(u_beats, v_beats):

        peaks, peak_ops, prom = find_two_peaks(vi.y)
        try:
            data["Maximum rise velocity"].append(vi.y[peaks[0]])
            data["Peak twitch amplitude"].append(np.max(ui.y))
            data["Maximum relaxation velocity"].append(vi.y[peaks[1]])
            data["Beat duration"].append(ui.t[-1] - ui.t[0])
            data["Time to peak twitch amplitude"].append(ui.ttp())
            data["Time to peak contraction velocity"].append(vi.t[peaks[0]] - vi.t[0])
            data["Time to peak relaxation velocity"].append(vi.t[peaks[1]] - vi.t[0])
            data["Width at half height"].append(ui.apd(50))
        except IndexError:
            continue

    return data



import functools
import logging
from typing import Optional
from typing import Union

try:
    from functools import cached_property  # type: ignore
except ImportError:
    # This is only supported in python 3.8 and above
    try:
        from cached_property import cached_property  # type: ignore
    except ImportError as e:
        raise ImportError(
            "Please install cached_property - pip install cached_property",
        ) from e

from dask.diagnostics import ProgressBar
import dask.array as da
import numpy as np
import dask
from scipy import interpolate

from . import frame_sequence as fs

Array = Union[da.Array, np.ndarray]
logger = logging.getLogger(__name__)


def compute_gradients(displacement: Array, dx=1) -> da.Array:
    """Compute gradients of the displacement

    Parameters
    ----------
    displacement : Array
        Displacement vectors
    dx : int, optional
        Number of difference steps for gradient computations, by default 1

    Returns
    -------
    da.Array
        Gradients
    """
    logger.info("Compute gradient using spline interpolation")
    shape = displacement.shape
    x = dx * np.arange(shape[0])
    y = dx * np.arange(shape[1])

    Uxs = []
    Uys = []
    for t in range(shape[2]):
        ux = displacement[:, :, t, 0]
        Uxs.append(dask.delayed(interpolate.RectBivariateSpline)(x, y, ux))
        uy = displacement[:, :, t, 1]
        Uys.append(dask.delayed(interpolate.RectBivariateSpline)(x, y, uy))

    dudxs = []
    dudys = []
    dvdxs = []
    dvdys = []

    logger.info("Compute interpolant Ux")
    with ProgressBar():
        Uxs = dask.compute(*Uxs)

    logger.info("Compute interpolant Uy")
    with ProgressBar():
        Uys = dask.compute(*Uys)

    for (
        Ux,
        Uy,
    ) in zip(Uxs, Uys):
        dudxs.append(dask.delayed(Ux)(x, y, dx=1).T)
        dudys.append(dask.delayed(Ux)(x, y, dy=1).T)
        dvdxs.append(dask.delayed(Uy)(x, y, dx=1).T)
        dvdys.append(dask.delayed(Uy)(x, y, dy=1).T)

    logger.info("Compute dudx")
    with ProgressBar():
        Dudx = da.stack(*dask.compute(dudxs))
    logger.info("Compute dudy")
    with ProgressBar():
        Dudy = da.stack(*dask.compute(dudys))
    logger.info("Compute dvdx")
    with ProgressBar():
        Dvdx = da.stack(*dask.compute(dvdxs))
    logger.info("Compute dvdy")
    with ProgressBar():
        Dvdy = da.stack(*dask.compute(dvdys))

    logger.info("Stack arrays")
    Du = da.stack([[Dudx, Dudy], [Dvdx, Dvdy]]).T
    logger.info("Done computing gradient")
    return Du


def compute_green_lagrange_strain_tensor(F: Array) -> da.Array:
    r"""Compute Green-Lagrange strain tensor

    .. math::

        \mathbf{E} = \frac{1}{2} \left( F^T F - I \right)

    Parameters
    ----------
    F : Array
        Deformation gradient

    Returns
    -------
    da.Array
        Green-Lagrange strain tensor
    """
    logger.debug("Compute Green Lagrange strain")
    F_t = da.transpose(F, (0, 1, 2, 4, 3))
    C = da.matmul(F_t, F)
    E = 0.5 * (C - da.eye(2)[None, None, None, :, :])

    return E


def compute_velocity(u: Array, t: Array, spacing: int = 1) -> da.Array:
    """Compute velocity from displacement

    Parameters
    ----------
    u : Array
        Displacement vectors
    t : Array
        Time stamps
    spacing : int, optional
        Number of steps between time steps to compute velocities, by default 1

    Returns
    -------
    da.Array
        Velocity
    """
    time_axis = u.shape.index(len(t))
    assert time_axis == 2, "Time axis should be the third axis"
    assert u.shape[-1] == 2, "Final axis should be ux and uy"
    assert spacing > 0, "Spacing must be a positive integer"

    if spacing == 1:
        dt = da.diff(t)
        du = da.diff(u, axis=time_axis)
    else:
        dt = t[spacing:] - t[:-spacing]
        du = u[:, :, spacing:, :] - u[:, :, :-spacing, :]

    # # Need to have time axis
    return da.moveaxis(da.moveaxis(du, 2, 3) / dt, 2, 3)


def compute_displacement(v: Array, t: Array, ref_index=0, spacing: int = 1) -> da.Array:
    """Compute displacement from velocity

    Parameters
    ----------
    v : Array
        Velocities
    t : Array
        time stamps
    ref_index : int, optional
        Index to be used as reference frame, by default 0
    spacing : int, optional
        Spacing used to compute velocities, by default 1

    Returns
    -------
    da.Array
        Displacement

    Raises
    ------
    NotImplementedError
        If spacing is different from 1
    """
    if spacing != 1:
        raise NotImplementedError("Only implemented for the case when spacing is 1")
    zero = da.zeros((v.shape[0], v.shape[1], 1, v.shape[3]))
    dt = np.diff(t)
    vdt = da.apply_along_axis(lambda x: x * dt, axis=2, arr=v)

    vdt_low = vdt[:, :, :ref_index, :]
    vdt_high = vdt[:, :, ref_index:, :]
    U = da.concatenate(
        (
            da.flip(da.cumsum(-da.flip(vdt_low), axis=2)),
            zero,
            da.cumsum(vdt_high, axis=2),
        ),
        axis=2,
    )

    return U


class Mechanics:
    def __init__(
        self,
        u: fs.VectorFrameSequence,
        t: Optional[Array] = None,
    ):
        """Create a mechanics object

        Parameters
        ----------
        u : fs.VectorFrameSequence
            Displacement of shape height x width x time x 2
        t : Optional[Array], optional
            Time stamps of length (time), by default None.
            If not provided `t` will be an evenly spaced
            array with a step of 1.0. Note that `t` is only
            relevant when computing time derivatives such as velocity.
        """
        assert isinstance(u, fs.VectorFrameSequence)
        self._u = u
        self.t = t

    @property
    def u(self) -> fs.VectorFrameSequence:
        """Displacement field"""
        return self._u

    @property
    def t(self) -> Array:
        """Time stamps"""
        return self._t

    @t.setter
    def t(self, t: Optional[Array]) -> None:
        if t is None:
            t = np.arange(self.num_time_points)
        if not len(t) == self.num_time_points:
            raise RuntimeError(
                "Expected time stamps to have the same number of points at 'u'",
            )
        self._t = t

    @property
    def dx(self) -> float:
        return self.u.dx

    @property
    def scale(self) -> float:
        return self.u.scale

    @property
    def height(self) -> int:
        return self.u.shape[0]

    @property
    def width(self) -> int:
        return self.u.shape[1]

    @property
    def num_time_points(self) -> int:
        return self.u.shape[2]

    def __str__(self):
        return (
            f"{self.__class__.__name__} object with "
            f"num_time_points={self.num_time_points}, "
            f"height={self.height}, "
            f"width={self.width}, "
            f"dx={self.dx}"
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(u={self.u}, dx={self.dx}, scale={self.scale})"
        )

    @cached_property
    def du(self) -> fs.TensorFrameSequence:

        try:
            du = compute_gradients(self.u.array, dx=self.dx)
        except ValueError:
            # We probably need to rechunk
            if isinstance(self.u._array, da.Array):
                logger.warning("Problems with computing gradient - try rechunking")
                self.u._array = self.u.array.rechunk()  # type:ignore
                du = compute_gradients(self.u.array, dx=self.dx)
            else:
                raise
        return fs.TensorFrameSequence(
            du,
            dx=self.dx,
            scale=self.scale,
        )

    @property
    def F(self) -> fs.TensorFrameSequence:
        """Deformation gradient"""
        return fs.TensorFrameSequence(
            self.du.array + da.eye(2)[None, None, None, :, :],
            dx=self.dx,
            scale=self.scale,
        )

    @property
    def E(self) -> fs.TensorFrameSequence:
        """Green-Lagrange strain tensor"""
        return fs.TensorFrameSequence(
            compute_green_lagrange_strain_tensor(self.F.array),
            dx=self.dx,
            scale=self.scale,
        )

    @functools.lru_cache
    def velocity(self, spacing: int = 1) -> fs.VectorFrameSequence:
        """Velocity field"""
        return fs.VectorFrameSequence(
            compute_velocity(self.u.array, self.t, spacing=spacing),
            dx=self.dx,
            scale=self.scale,
        )

    @cached_property
    def principal_strain(self) -> fs.VectorFrameSequence:
        """Return the principal strain

        Parameters
        ----------
        k : int, optional
            Which component of the princial strain either 0 or 1, by default 0.
            0 correspond to the largets eigen value
        recompute : bool, optional
            If already computed set this to True if you want to recompute, by default False

        Returns
        -------
        np.ndarray
            The kth principal strain
        """
        return self.E.compute_eigenvalues()



__doc__ = """
Motion tracking of MPS data
This is software to estimate motion in Brightfield images.
"""
import datetime
import logging
from pathlib import Path
from typing import Any, Dict, Optional

import matplotlib.pyplot as plt
import json

from . import Mechanics, OpticalFlow
from . import motion_tracking as mt
from . import utils
from . import stats
from . import visu
from . import scaling

logger = logging.getLogger(__name__)


def print_dict(d: Dict[str, Any], fmt="{:<10}: {}"):
    s = ""
    for k, v in d.items():
        s += fmt.format(k, str(v)) + "\n"
    return s


def plot_traces(results, outdir: Path, time_unit="ms"):
    assert "time" in results, "Missing time array"
    for name, arr in results.items():
        if name == "time":
            continue

        fig, ax = plt.subplots()
        ax.plot(results["time"][: len(arr)], arr)
        ax.set_xlabel(f"Time [{time_unit}]")
        if name.startswith("u"):
            ylabel = f"Displacement {name.split('_')[-1]} [\u00B5m]"
        else:
            ylabel = f"Velocity {name.split('_')[-1]} [\u00B5m/s]"
        ax.set_ylabel(ylabel)
        ax.grid()
        fig.savefig(outdir.joinpath(name).with_suffix(".png"))


def load_data(filename_):
    try:
        import mps

        data = mps.MPS(filename_)
    except ImportError as e:
        raise ImportError(
            (
                "Missing `mps` pacakge. Please install it. "
                "'python -m pip install cardiac-mps'"
            ),
        ) from e
    return data


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Path):
            return obj.as_posix()
        elif isinstance(mt.FLOW_ALGORITHMS):
            return str(obj)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def main(
    filename: str,
    algorithm: mt.FLOW_ALGORITHMS = mt.FLOW_ALGORITHMS.farneback,
    outdir: Optional[str] = None,
    reference_frame: str = "0",
    scale: float = 1.0,
    apply_filter: bool = True,
    spacing: int = 5,
    compute_xy_components: bool = False,
    make_displacement_video: bool = True,
    make_velocity_video: bool = True,
    verbose: bool = False,
):
    """
    Estimate motion in stack of images

    Parameters
    ----------
    filename : str
        Path to file to be analyzed, typically an .nd2 or .czi Brightfield file
    algoritm : mt.FLOW_ALGORITHMS, optional
        The algorithm used to estimate motion, by default mt.FLOW_ALGORITHMS.farneback
    outdir : Optional[str], optional
        Directory where to store the results. If not provided, a folder with the the same
        as the filename will be created and results will be stored in a subfolder inside
        that called `motion`.
    scale : float, optional
        Rescale data before running motion track. This is useful if the spatial resoltion
        of the images are large. Scale = 1.0 will keep the original size, by default 0.3
    apply_filter, bool, optional
        If True, set pixels with max displacement lower than the mean maximum displacement
        to zero. This will prevent non-tissue pixels to be included, which is especially
        error prone for velocity estimations, by default True.
    spacing: int, optional
        Spacing between frames in velocity computations, by default 5.
    compute_xy_components: bool, optional
        If True the compute x- and y components of the displacement and
        velocity and plot them as well, by default False.
    make_displacement_video: bool, optional
        If True, create video of displacement vectors, by default False.
    make_velocity_video: bool, optional
        If True, create video of velocity vectors, by default False.
    verbose : bool, optional
        Print more to the console, by default False
    overwrite : bool, optional
        If `outdir` allready exist an contains the relevant files then set this to false to
        use that data, by default True

    Raises
    ------
    IOError
        If filename does not exist
    ValueError
        If scale is zero or lower, or higher than 1.0
    """
    if verbose:
        logger.setLevel(logging.DEBUG)
        utils.logger.setLevel(logging.DEBUG)
        mt.logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
        utils.logger.setLevel(logging.INFO)
        mt.logger.setLevel(logging.INFO)

    filename_ = Path(filename)
    if outdir is None:
        outdir_ = filename_.with_suffix("").joinpath("motion")
    else:
        outdir_ = Path(outdir)

    settings = {
        "filename": filename_,
        "algorithm": algorithm,
        "outdir": outdir_,
        "scale": scale,
        "reference_frame": reference_frame,
        "timestamp": datetime.datetime.now().isoformat(),
    }
    logger.debug("\nSettings : \n{}".format(print_dict(settings)))

    settings_file = outdir_.joinpath("settings.json")
    results_file = outdir_.joinpath("results.csv")
    features_file = outdir_.joinpath("features.csv")

    if not filename_.is_file():
        raise IOError(f"File {filename_} does not exist")

    if not (0 < scale <= 1.0):
        raise ValueError("Scale has to be between 0 and 1.0")

    data = load_data(filename_)
    logger.info(f"Analyze motion in file {filename}...")
    if scale < 1.0:
        data = scaling.resize_data(data, scale=scale)
    opt_flow = OpticalFlow(
        data,
        flow_algorithm=algorithm,
    )
    u = opt_flow.get_displacements(reference_frame=reference_frame)
    factor = 1000.0 if data.info["time_unit"] == "ms" else 1.0
    v = Mechanics(u, t=data.time_stamps / factor).velocity(spacing=spacing)
    if apply_filter:
        logger.info("Apply filter")
        u_norm_max = u.norm().max().compute()
        mask = u_norm_max < u_norm_max.mean()
        v.apply_mask(mask)
        u.apply_mask(mask)

    results = {"time": data.time_stamps}
    logger.info("Compute displacement norm")
    results["u_norm"] = u.norm().mean().compute()
    logger.info("Compute velocity norm")
    results["v_norm"] = v.norm().mean().compute()

    if compute_xy_components:
        logger.info("Compute displacement x component")
        results["u_x"] = u.x.mean().compute()
        logger.info("Compute displacement y component")
        results["u_y"] = u.y.mean().compute()
        logger.info("Compute velocity x component")
        results["v_x"] = v.x.mean().compute()
        logger.info("Compute velocity y component")
        results["v_y"] = v.y.mean().compute()

    outdir_.mkdir(exist_ok=True, parents=True)
    plot_traces(results=results, outdir=outdir_, time_unit=data.info["time_unit"])

    logger.info("Compute features")
    features = stats.compute_features(
        u=results["u_norm"],
        v=results["v_norm"],
        t=data.time_stamps,
    )
    with open(settings_file, "w") as f:
        json.dump(settings, f, cls=JSONEncoder, indent=2)

    import mps

    mps.utils.to_csv(results, results_file)
    mps.utils.to_csv(features, features_file)

    if make_displacement_video:
        visu.quiver_video(
            data,
            u,
            outdir_.joinpath("displacement_movie"),
            step=24,
            vector_scale=4,
        )

    if make_velocity_video:
        visu.quiver_video(
            data,
            v,
            outdir_.joinpath("velocity_movie"),
            step=24,
            vector_scale=1,
            offset=spacing,
        )



"""Top-level package for MPS Motion Tracking."""
import logging as _logging
from importlib.metadata import metadata

import daiquiri as _daiquiri

from . import block_matching
from . import dualtvl1
from . import farneback
from . import filters
from . import frame_sequence
from . import lucas_kanade
from . import mechanics
from . import motion_tracking
from . import scaling
from . import stats
from . import utils
from . import visu
from .frame_sequence import FrameSequence
from .frame_sequence import TensorFrameSequence
from .frame_sequence import VectorFrameSequence
from .mechanics import Mechanics
from .motion_tracking import FLOW_ALGORITHMS
from .motion_tracking import list_optical_flow_algorithm
from .motion_tracking import OpticalFlow
from .utils import MPSData

meta = metadata("mps-motion")
__version__ = meta["Version"]
__author__ = meta["Author"]
__license__ = meta["License"]
__email__ = meta["Author-email"]
__program_name__ = meta["Name"]


def set_log_level(level):
    from daiquiri import set_default_log_levels

    loggers = [
        "block_matching.logger",
        "dualtvl1.logger",
        "farneback.logger",
        "lucas_kanade.logger",
        "mechanics.logger",
        "motion_tracking.logger",
        "scaling.logger",
        "utils.logger",
        "visu.logger",
        "frame_sequence.logger",
        "filters.logger",
    ]
    set_default_log_levels((logger, level) for logger in loggers)


_daiquiri.setup(level=_logging.INFO)

__all__ = [
    "farneback",
    "dualtvl1",
    "lucas_kanade",
    "block_matching",
    "utils",
    "mechanics",
    "frame_sequence",
    "Mechanics",
    "MPSData",
    "motion_tracking",
    "FLOW_ALGORITHMS",
    "OpticalFlow",
    "scaling",
    "frame_sequence",
    "visu",
    "filters",
    "FrameSequence",
    "VectorFrameSequence",
    "TensorFrameSequence",
    "stats",
    "list_optical_flow_algorithm",
]



from mps_motion.__main__ import app
from typer.testing import CliRunner


def test_cli(TEST_FILENAME):
    runner = CliRunner()
    result = runner.invoke(app, ["analyze", TEST_FILENAME])
    assert result.exit_code == 0



import logging
import os
from typing import Union

import dask.array as da
import numpy as np


logger = logging.getLogger(__name__)

try:
    from numba import jit, prange
except ImportError:
    msg = (
        "numba not found - Numba is just to speed up the motion tracking algorithm\n"
        "To install numba use: pip install numba"
    )
    logger.debug(msg)
    prange = range

    # Create a dummy decorator
    def jit(**params):
        def decorator(f):
            logger.warning(
                "You are trying to call a numba function, but numba is not installed",
            )

            def wrap(*args, **kwargs):
                return f(*args, **kwargs)

            return wrap

        return decorator


class ShapeError(RuntimeError):
    pass


def download_demo_data(path):
    print("Downloading data. Please wait...")
    link = "https://www.dropbox.com/s/lmo980fzsvxnptj/data.npy?dl=1"
    import urllib.request
    import time

    urllib.request.urlretrieve(link, path)
    time.sleep(1.0)
    print("Done downloading data")


def unmask(array, fill_value=0.0):
    """Take a mask array and fills
    it with the given fill value and
    return an regular numpy array
    """
    if isinstance(array, np.ma.MaskedArray):
        return array.filled(fill_value)
    return array


PathLike = Union[str, os.PathLike]
Array = Union[da.core.Array, np.ndarray]


def check_frame_dimensions(frames, reference_image):
    if not isinstance(frames, np.ndarray):
        frames = np.asanyarray(frames)
    if len(frames.shape) != 3:
        raise ShapeError(f"Expected frame to be 3 dimensional, got {frames.shape}")
    if len(reference_image.shape) != 2:
        raise ShapeError(
            f"Expected refernce image to be 2 dimensional, got {frames.shape}",
        )
    num_frames = frames.shape[-1]
    shape = reference_image.shape
    if frames.shape != (shape[0], shape[1], num_frames):
        msg = (
            "Shape mistmact between frames and reference image. "
            f"Got frames with shape {frames.shape}, and reference "
            f"image with shape {reference_image.shape}"
        )
        raise ShapeError(msg)

    return frames


class MPSData:
    def __init__(self, frames, time_stamps, info, pacing=None, metadata=None) -> None:
        self.frames = frames
        self.time_stamps = time_stamps
        self.info = info
        if pacing is None:
            pacing = np.zeros_like(time_stamps)
        self.pacing = pacing
        if metadata is None:
            metadata = {}
        self.metadata = metadata

    @property
    def size_x(self):
        return self.frames.shape[0]

    @property
    def size_y(self):
        return self.frames.shape[1]

    @property
    def num_frames(self):
        return self.frames.shape[2]

    @property
    def framerate(self):
        return int(1000 / np.mean(np.diff(self.time_stamps)))


def to_uint8(img):
    img_float = img.astype(float)
    return (256 * (img_float / max(img_float.max(), 1e-12))).astype(np.uint8)


def ca_transient(
    t: np.ndarray,
    tstart: float = 0.05,
    tau1: float = 0.05,
    tau2: float = 0.110,
    ca_diast: float = 0.0,
    ca_ampl: float = 1.0,
) -> np.ndarray:

    beta = (tau1 / tau2) ** (-1 / (tau1 / tau2 - 1)) - (tau1 / tau2) ** (
        -1 / (1 - tau2 / tau1)
    )
    ca = np.zeros_like(t)

    ca[t <= tstart] = ca_diast

    ca[t > tstart] = (ca_ampl - ca_diast) / beta * (
        np.exp(-(t[t > tstart] - tstart) / tau1)
        - np.exp(-(t[t > tstart] - tstart) / tau2)
    ) + ca_diast
    return ca



import logging
from concurrent.futures import ProcessPoolExecutor
from enum import Enum
from typing import List
from typing import Optional

import dask
import dask.array as da
import dask_image.ndfilters
import numpy as np
import tqdm
from dask.diagnostics import ProgressBar
from scipy.interpolate import bisplev
from scipy.interpolate import bisplrep
from typing_extensions import Protocol

from . import utils

logger = logging.getLogger(__name__)


def spline_smooth_u(args):
    ux, uy, X, Y = args

    N = min(ux.shape) // 10

    x = X.flatten()[::N]
    y = Y.flatten()[::N]
    valuesx = ux.flatten()[::N]
    valuesy = uy.flatten()[::N]

    sx = X.shape[1] - np.sqrt(2 * X.shape[1])
    sy = X.shape[0] - np.sqrt(2 * X.shape[0])

    tckx = bisplrep(x, y, valuesx, kx=5, ky=5, s=sx)
    Ux = bisplev(X[:, 0], Y[0, :], tckx)
    tcky = bisplrep(x, y, valuesy, kx=5, ky=5, s=sy)
    Uy = bisplev(X[:, 0], Y[0, :], tcky)

    return [Ux, Uy]


def process_spline_interpolation(u):
    x = np.arange(u.shape[0])
    y = np.arange(u.shape[1])
    Y, X = np.meshgrid(y, x)

    Us = []

    args = ((u[:, :, i, 0], u[:, :, i, 1], X, Y) for i in range(u.shape[2]))

    with ProcessPoolExecutor() as executor:
        for ui in tqdm.tqdm(
            executor.map(spline_smooth_u, args),
            desc="Running spline interpolation",
            total=u.shape[2],
        ):
            Us.append(ui)

    return Us


def spline_smooth(u: utils.Array) -> np.ndarray:

    if isinstance(u, da.Array):
        u = u.compute()

    logger.info("Performing spline interpolation, this may take some time...")
    Us = process_spline_interpolation(u)

    new_u = np.zeros(u.shape)

    for i, ui in enumerate(Us):
        new_u[:, :, i, 0] = ui[0]
        new_u[:, :, i, 1] = ui[1]

    return new_u


class Filters(str, Enum):
    median = "median"
    gaussian = "gaussian"


def filter_vectors(
    vectors: utils.Array,
    filter_type: Filters,
    size: Optional[int] = None,
    sigma: Optional[float] = None,
):

    if not valid_filter(filter_type=filter_type, size=size, sigma=sigma):
        return vectors

    is_numpy = False
    if isinstance(vectors, np.ndarray):
        is_numpy = True
        vectors = da.from_array(vectors)

    vec0 = apply_filter(
        vectors[:, :, 0],
        filter_type=filter_type,
        size=size,
        sigma=sigma,
    )
    vec1 = apply_filter(
        vectors[:, :, 1],
        filter_type=filter_type,
        size=size,
        sigma=sigma,
    )
    vectors = da.stack([vec0, vec1], axis=-1)
    if is_numpy:
        vectors = vectors.compute()
    return vectors


def filter_vectors_par(
    vectors,
    filter_type: Filters,
    size: Optional[int] = None,
    sigma: Optional[float] = None,
):

    if not valid_filter(filter_type=filter_type, size=size, sigma=sigma):
        return vectors
    logger.info("Filter vectors")
    assert len(vectors.shape) == 4
    assert vectors.shape[3] == 2
    num_frames = vectors.shape[2]

    is_numpy = False
    if isinstance(vectors, np.ndarray):
        is_numpy = True
        vectors = da.from_array(vectors)

    all_vectors = []
    for i in range(num_frames):
        all_vectors.append(
            dask.delayed(filter_vectors)(
                vectors[:, :, i, :],
                filter_type=filter_type,
                size=size,
                sigma=sigma,
            ),
        )

    with ProgressBar():
        vectors = da.stack(*da.compute(all_vectors), axis=2)
    if is_numpy:
        vectors = vectors.compute()

    logger.info("Done filtering")
    return vectors


def valid_filter(
    filter_type: Filters,
    size: Optional[int] = None,
    sigma: Optional[float] = None,
) -> bool:
    if filter_type == Filters.median and size is None:
        logger.warning("Please provide a size of the median filter kernel")
        return False
    elif filter_type == Filters.gaussian and sigma is None:
        logger.warning("Please provide a sigma for the gaussian filter")
        return False
    if filter_type not in Filters._member_names_:
        logger.warning(f"Unknown filter type {filter_type}")
        return False
    return True


def is_positive(value: Optional[float]) -> bool:
    if value is None:
        return False
    return value > 0


def apply_filter(
    array: utils.Array,
    filter_type: Filters,
    size: Optional[int] = None,
    sigma: Optional[float] = None,
) -> utils.Array:
    if not valid_filter(filter_type=filter_type, size=size, sigma=sigma):
        return array

    is_numpy = False
    if isinstance(array, np.ndarray):
        is_numpy = True
        array = da.from_array(array)

    if filter_type == Filters.median and is_positive(size):
        array = dask_image.ndfilters.median_filter(array, size)

    elif filter_type == Filters.gaussian and is_positive(sigma):
        array = dask_image.ndfilters.gaussian_filter(array, sigma)

    if is_numpy:
        array = array.compute()

    return array


class InvalidThresholdError(ValueError):
    pass


def check_threshold(
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
):
    if (vmax and vmin) and vmax < vmin:
        raise InvalidThresholdError(f"Cannot have vmax < vmin, got {vmax=} and {vmin=}")


def threshold(
    array: utils.Array,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    copy: bool = True,
) -> utils.Array:
    """Threshold an array

    Parameters
    ----------
    array : utils.Array
        The array
    vmin : Optional[float], optional
        Lower threshold value, by default None
    vmax : Optional[float], optional
        Upper threshold value, by default None
    copy : bool, optional
        Operative on the given array or use a copy, by default True

    Returns
    -------
    utils.Array
        Inpute array with the lowest value beeing vmin and
        highest value begin vmax
    """
    assert len(array.shape) == 3
    check_threshold(vmin, vmax)
    if copy:
        array = array.copy()

    if vmax is not None:
        array[array > vmax] = vmax
    if vmin is not None:
        array[array < vmin] = vmin
    return array


def _handle_threshold_norm(norm_inds, ns, factor, norm_array, array):
    if norm_inds.any():
        if ns == da:
            norm_inds = norm_inds.compute()
        inds = np.stack([norm_inds, norm_inds], -1).flatten()
        values = (
            factor
            / ns.stack([norm_array[norm_inds], norm_array[norm_inds]], -1).flatten()
        )
        array[inds] *= values


class _Linalg(Protocol):
    @staticmethod
    def norm(array: utils.Array, axis: int = 0) -> utils.Array:
        ...


class NameSpace(Protocol):
    @property
    def linalg(self) -> _Linalg:
        ...

    @staticmethod
    def stack(arrs: List[utils.Array], axis: int) -> utils.Array:
        ...


def threshold_norm(
    array: utils.Array,
    ns: NameSpace,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    copy: bool = True,
) -> utils.Array:
    """Threshold an array of vectors based on the
    norm of the vectors.

    For example if the vectors are displacement then
    you can use this function to scale all vectors so
    that the magnitudes are within `vmin` and `vmax`

    Parameters
    ----------
    array : utils.Array
        The input array which is 4D and the last dimension is 2.
    ns : NameSpace
        Wheter to use numpy or dask
    vmin : Optional[float], optional
        Lower bound on the norm, by default None
    vmax : Optional[float], optional
        Upper bound on the norm, by default None
    copy : bool, optional
        Wheter to operate on the input are or use a copy, by default True

    Returns
    -------
    utils.Array
        The thresholded array
    """
    assert len(array.shape) == 4
    assert array.shape[3] == 2
    assert ns in [da, np]
    check_threshold(vmin, vmax)
    if copy:
        array = array.copy()
    shape = array.shape
    norm_array = ns.linalg.norm(array, axis=3).flatten()
    array = array.flatten()

    if vmax is not None:
        norm_inds = norm_array > vmax
        _handle_threshold_norm(norm_inds, ns, vmax, norm_array, array)
    if vmin is not None:
        norm_inds = norm_array < vmin
        _handle_threshold_norm(norm_inds, ns, vmin, norm_array, array)
    return array.reshape(shape)
